@@ -8,15 +8,16 @@
 import sys
 import logging
 from typing import Any
+from collections.abc import Sequence
 
-from xpra.util.str_fn import csv, print_nested_dict, bytestostr
+from xpra.util.str_fn import csv, print_nested_dict
 from xpra.util.env import envint, envbool, numpy_import_context
-from xpra.log import Logger, CaptureHandler
+from xpra.log import Logger, CaptureHandler, consume_verbose_argv
 from xpra.client.gl.drivers import GL_MATCH_LIST, WHITELIST, GREYLIST, BLACKLIST, OpenGLFatalError
 
 log = Logger("opengl")
 
-required_extensions: tuple[str, ...] = ("GL_ARB_texture_rectangle", "GL_ARB_vertex_program")
+required_extensions: Sequence[str] = ("GL_ARB_texture_rectangle", "GL_ARB_vertex_program")
 
 GL_ALPHA_SUPPORTED: bool = envbool("XPRA_ALPHA", True)
 DOUBLE_BUFFERED: bool = envbool("XPRA_OPENGL_DOUBLE_BUFFERED", True)
@@ -73,7 +74,7 @@ def get_max_texture_size() -> int:
     return int(min(rect_texture_size, texture_size))
 
 
-def get_array_handlers() -> tuple[str, ...]:
+def get_array_handlers() -> Sequence[str]:
     from OpenGL.arrays.arraydatatype import ArrayDatatype
     try:
         return tuple(getattr(atype, "__name__", str(atype)) for atype in set(ArrayDatatype.getRegistry().values()))
@@ -179,7 +180,7 @@ def get_GLU_info() -> dict[str, str]:
 
 
 def get_context_info() -> dict[str, Any]:
-    with numpy_import_context("OpenGL", True):
+    with numpy_import_context("OpenGL: context info", True):
         from OpenGL.GL import glGetIntegerv
         from OpenGL.GL import (
             GL_CONTEXT_PROFILE_MASK, GL_CONTEXT_CORE_PROFILE_BIT,
@@ -227,21 +228,21 @@ def check_base_functions() -> str:
         glActiveTexture, glTexSubImage2D, glTexCoord2i,
         glViewport, glMatrixMode, glLoadIdentity, glOrtho,
         glEnableClientState, glGenTextures, glDisable,
-        glBindTexture, glPixelStorei, glEnable, glBegin, glFlush,
+        glBindTexture, glPixelStorei, glEnable, glFlush,
         glTexParameteri, glTexEnvi, glHint, glBlendFunc, glLineStipple,
         glTexImage2D,
         glMultiTexCoord2i,
-        glVertex2i, glEnd,
+        glVertex2i,
     )
     return check_available(
         glActiveTexture, glTexSubImage2D, glTexCoord2i,
         glViewport, glMatrixMode, glLoadIdentity, glOrtho,
         glEnableClientState, glGenTextures, glDisable,
-        glBindTexture, glPixelStorei, glEnable, glBegin, glFlush,
+        glBindTexture, glPixelStorei, glEnable, glFlush,
         glTexParameteri, glTexEnvi, glHint, glBlendFunc, glLineStipple,
         glTexImage2D,
         glMultiTexCoord2i,
-        glVertex2i, glEnd,
+        glVertex2i,
     )
 
 
@@ -318,14 +319,16 @@ def check_PyOpenGL_support(force_enable) -> dict[str, Any]:
         if TIMEOUT > 0:
             import time
             time.sleep(TIMEOUT)
+
         # log redirection:
         for name in ("formathandler", "extensions", "acceleratesupport", "arrays", "converters", "plugins"):
             logger = logging.getLogger(f"OpenGL.{name}")
             redirected_loggers[name] = (logger, list(logger.handlers), logger.propagate)
             logger.handlers = [CaptureHandler()]
             logger.propagate = False
+        log(f"{redirected_loggers=}")
 
-        with numpy_import_context("OpenGL", True):
+        with numpy_import_context("OpenGL: check", True):
             return do_check_PyOpenGL_support(force_enable)
 
     finally:
@@ -350,6 +353,7 @@ def check_PyOpenGL_support(force_enable) -> dict[str, Any]:
         strip_log_message = "Unable to load registered array format handler "
         missing_handlers = []
         for msg in recs("formathandler"):
+            log.info(f"msg formathandler={msg}")
             p = msg.find(strip_log_message)
             if p < 0:
                 # unknown message, log it:
@@ -364,9 +368,10 @@ def check_PyOpenGL_support(force_enable) -> dict[str, Any]:
             log.warn("PyOpenGL warning: missing array format handlers: %s", csv(missing_handlers))
 
         for msg in recs("extensions"):
-            # ignore extension messages:
-            p = msg.startswith("GL Extension ") and msg.endswith("available")
-            if not p:
+            # hide extension messages:
+            if msg.startswith("GL Extension ") or msg.endswith("available"):
+                log(msg)
+            else:
                 log.info(msg)
 
         missing_accelerators = []
@@ -383,7 +388,11 @@ def check_PyOpenGL_support(force_enable) -> dict[str, Any]:
             else:
                 log.info(msg)
         if missing_accelerators:
-            log.info("OpenGL accelerate missing: %s", csv(missing_accelerators))
+            if len(missing_accelerators) == 1 and missing_accelerators[0] == "numpy_formathandler":
+                log_fn = log.debug
+            else:
+                log_fn = log.info
+            log_fn("OpenGL accelerate missing: %s", csv(missing_accelerators))
 
         for msg in recs("plugins"):
             log(f"plugins msg={msg}")
@@ -413,18 +422,18 @@ def do_check_PyOpenGL_support(force_enable) -> dict[str, Any]:
     import OpenGL
     props["pyopengl"] = OpenGL.__version__  # @UndefinedVariable
     from OpenGL.GL import GL_VERSION, glGetString
-    gl_version_str = glGetString(GL_VERSION)
-    if gl_version_str is None and not force_enable:
+    gl_version_str = (glGetString(GL_VERSION) or b"").decode("latin1")
+    if not gl_version_str and not force_enable:
         raise_fatal_error("OpenGL version is missing - cannot continue")
         return props
-    # b'4.6.0 NVIDIA 440.59' -> ['4', '6', '0 NVIDIA...']
-    log("GL_VERSION=%s", bytestostr(gl_version_str))
-    vparts = bytestostr(gl_version_str).split(" ", 1)[0].split(".")
+    # '4.6.0 NVIDIA 440.59' -> ['4', '6', '0 NVIDIA...']
+    log("GL_VERSION=%s", gl_version_str)
+    vparts = gl_version_str.split(" ", 1)[0].split(".")
     try:
         gl_major = int(vparts[0])
         gl_minor = int(vparts[1])
     except (IndexError, ValueError) as e:
-        msg = "failed to parse gl version '%s': %s" % (bytestostr(gl_version_str), e)
+        msg = "failed to parse gl version '%s': %s" % (gl_version_str, e)
         unsafe(msg)
         log(" assuming this is at least 1.1 to continue")
     else:
@@ -465,7 +474,7 @@ def do_check_PyOpenGL_support(force_enable) -> dict[str, Any]:
         if not force_enable:
             raise_fatal_error(msg)
             return props
-    props["zerocopy"] = accel_version and pyopengl_version == accel_version
+    props["zerocopy"] = bool(accel_version and pyopengl_version == accel_version)
 
     props.update(get_vendor_info())
     props.update(get_GLU_info())
@@ -491,6 +500,7 @@ def do_check_PyOpenGL_support(force_enable) -> dict[str, Any]:
     safe = check_lists(props, force_enable)
     if "safe" not in props:
         props["safe"] = safe
+    props.update(get_context_info())
     return props
 
 
@@ -502,9 +512,7 @@ def main() -> int:
     with program_context("OpenGL-Check"):
         gui_init()
         enable_color()
-        verbose = "-v" in sys.argv or "--verbose" in sys.argv
-        if verbose:
-            log.enable_debug()
+        consume_verbose_argv(sys.argv, "opengl")
         from xpra.gtk.util import init_display_source
         init_display_source()
         force_enable = "-f" in sys.argv or "--force" in sys.argv
@@ -527,7 +535,6 @@ def main() -> int:
         gl_fatal_error = log_error
         try:
             props = gl_context.check_support(force_enable)
-            props.update(get_context_info())
         except Exception as e:
             props = {}
             log("check_support", exc_info=True)
