@@ -8,14 +8,14 @@
 import os
 import weakref
 from time import monotonic
-from collections.abc import Callable
+from collections.abc import Callable, Sequence, Iterable
 from subprocess import Popen, PIPE
 from threading import Event
 from typing import Any
 
 from xpra.common import noop
 from xpra.util.objects import typedict
-from xpra.util.str_fn import csv, ellipsizer, repr_ellipsized, pver, strtobytes, bytestostr, hexstr, memoryview_to_bytes
+from xpra.util.str_fn import csv, Ellipsizer, repr_ellipsized, pver, strtobytes, bytestostr, hexstr, memoryview_to_bytes
 from xpra.util.env import envint, envbool, osexpand, first_time, IgnoreWarningsContext, ignorewarnings
 from xpra.util.child_reaper import getChildReaper
 from xpra.os_util import gi_import, WIN32, OSX, POSIX
@@ -29,7 +29,7 @@ from xpra.gtk.cursors import cursor_types, get_default_cursor
 from xpra.gtk.util import get_default_root_window, get_root_size, GRAB_STATUS_STRING, init_display_source
 from xpra.gtk.window import GDKWindow
 from xpra.gtk.info import get_monitors_info, get_screen_sizes
-from xpra.gtk.widget import scaled_image, label, color_parse
+from xpra.gtk.widget import scaled_image, label, color_parse, FILE_CHOOSER_NATIVE
 from xpra.gtk.pixbuf import get_icon_pixbuf, get_pixbuf_from_data
 from xpra.gtk.versions import get_gtk_version_info
 from xpra.exit_codes import ExitCode, ExitValue
@@ -58,6 +58,7 @@ notifylog = Logger("gtk", "notify")
 grablog = Logger("client", "grab")
 focuslog = Logger("client", "focus")
 
+GLib = gi_import("GLib")
 Gtk = gi_import("Gtk")
 Gdk = gi_import("Gdk")
 GdkPixbuf = gi_import("GdkPixbuf")
@@ -75,6 +76,7 @@ NO_OPENGL_WINDOW_TYPES = os.environ.get(
     "XPRA_NO_OPENGL_WINDOW_TYPES",
     "DOCK,TOOLBAR,MENU,UTILITY,SPLASH,DROPDOWN_MENU,POPUP_MENU,TOOLTIP,NOTIFICATION,COMBO,DND"
 ).split(",")
+WINDOW_GROUPING = os.environ.get("XPRA_WINDOW_GROUPING", "group-leader-xid,class-instance,pid,command").split(",")
 
 inject_css_overrides()
 init_display_source()
@@ -82,6 +84,16 @@ init_display_source()
 from xpra.client.gtk3.window_base import HAS_X11_BINDINGS, XSHAPE  # noqa: E402
 
 
+def get_group_ref(metadata: typedict) -> str:
+    for ref in WINDOW_GROUPING:
+        if ref in metadata:
+            value = metadata.get(ref)
+            if isinstance(value, Iterable):
+                return f"{ref}:{csv(value)}"
+            return f"{ref}:{value}"
+    return ""
+
+
 # pylint: disable=import-outside-toplevel
 def show_border_help() -> None:
     if not first_time("border-help"):
@@ -129,6 +141,7 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
         self.gl_max_viewport_dims = 0, 0
         self.gl_texture_size_limit = 0
         self._cursors = weakref.WeakKeyDictionary()
+        self.border_str = "no"
         # frame request hidden window:
         self.frame_request_window = None
         # group leader bits:
@@ -147,6 +160,7 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
     def init(self, opts) -> None:
         GObjectXpraClient.init(self, opts)
         UIXpraClient.init(self, opts)
+        self.border_str = opts.border
 
     def setup_frame_request_windows(self) -> None:
         # query the window manager to get the frame size:
@@ -165,9 +179,11 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
     def run(self) -> ExitValue:
         log(f"run() HAS_X11_BINDINGS={HAS_X11_BINDINGS}")
         # call this once early:
-        ignorewarnings(self.get_mouse_position)
-        if HAS_X11_BINDINGS:
-            self.setup_frame_request_windows()
+        from xpra.client.gui import features
+        if features.windows:
+            ignorewarnings(self.get_mouse_position)
+            if HAS_X11_BINDINGS:
+                self.setup_frame_request_windows()
         UIXpraClient.run(self)
         self.gtk_main()
         log(f"GTKXpraClient.run_main_loop() main loop ended, returning exit_code={self.exit_code}", )
@@ -184,14 +200,14 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
             self.exit_code = exit_code
         if Gtk.main_level() > 0:
             # if for some reason cleanup() hangs, maybe this will fire...
-            self.timeout_add(4 * 1000, self.exit)
+            GLib.timeout_add(4 * 1000, self.exit)
             # try harder!:
-            self.timeout_add(5 * 1000, self.force_quit, exit_code)
+            GLib.timeout_add(5 * 1000, self.force_quit, exit_code)
         self.cleanup()
         log(f"GTKXpraClient.quit({exit_code}) cleanup done, main_level={Gtk.main_level()}")
         if Gtk.main_level() > 0:
             log(f"GTKXpraClient.quit({exit_code}) main loop at level {Gtk.main_level()}, calling gtk quit via timeout")
-            self.timeout_add(500, self.exit)
+            GLib.timeout_add(500, self.exit)
 
     def exit(self) -> None:
         self.show_progress(100, "terminating")
@@ -228,13 +244,13 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
 
     def start_UI_watcher(self, _client) -> None:
         from xpra.platform.ui_thread_watcher import get_UI_watcher
-        self.UI_watcher = get_UI_watcher(self.timeout_add, self.source_remove)
+        self.UI_watcher = get_UI_watcher()
         assert self.UI_watcher
         self.UI_watcher.start()
 
         # if server supports it, enable UI thread monitoring workaround when needed:
 
-        def UI_resumed():
+        def UI_resumed() -> None:
             self.send("resume", True, tuple(self._id_to_window.keys()))
             # maybe the system was suspended?
             # so we may want to call WindowClient.resume()
@@ -242,7 +258,7 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
             if resume:
                 resume()
 
-        def UI_failed():
+        def UI_failed() -> None:
             self.send("suspend", True, tuple(self._id_to_window.keys()))
 
         self.UI_watcher.add_resume_callback(UI_resumed)
@@ -271,6 +287,7 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
         Gdk.notify_startup_complete()
 
     def do_process_challenge_prompt(self, prompt="password"):
+        self.stop_progress_process(f"showing {prompt} prompt")
         authlog = Logger("auth")
         self.show_progress(100, "authentication")
         PINENTRY = os.environ.get("XPRA_PINENTRY", "")
@@ -325,10 +342,10 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
         values = []
         errs = []
 
-        def rec(value=None):
+        def rec(value=None) -> None:
             values.append(value)
 
-        def err(value=None):
+        def err(value=None) -> None:
             errs.append(value)
 
         from xpra.scripts.pinentry import pinentry_getpin
@@ -346,7 +363,7 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
         # so we block the current thread using an event:
         wait = Event()
         values = []
-        self.idle_add(self.do_process_challenge_prompt_dialog, values, wait, prompt)
+        GLib.idle_add(self.do_process_challenge_prompt_dialog, values, wait, prompt)
         wait.wait()
         if not values:
             return None
@@ -363,7 +380,7 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
         dialog.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.REJECT)
         dialog.add_button(Gtk.STOCK_OK, Gtk.ResponseType.ACCEPT)
 
-        def add(widget, padding=0):
+        def add(widget, padding=0) -> None:
             widget.set_halign(Gtk.Align.CENTER)
             widget.set_margin_start(padding)
             widget.set_margin_end(padding)
@@ -382,7 +399,7 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
         dialog.vbox.show_all()
         dialog.password_input = password_input
 
-        def handle_response(dialog, response):
+        def handle_response(dialog, response) -> None:
             if values:
                 return
             if OSX:
@@ -404,7 +421,7 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
                 values.append(password)
             wait.set()
 
-        def password_activate(*_args):
+        def password_activate(*_args) -> None:
             handle_response(dialog, Gtk.ResponseType.ACCEPT)
 
         password_input.connect("activate", password_activate)
@@ -509,10 +526,12 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
 
     ################################
     # file handling
-    def ask_data_request(self, cb_answer, send_id, dtype: str, url: str, filesize: int, printit: bool, openit: bool):
-        self.idle_add(self.do_ask_data_request, cb_answer, send_id, dtype, url, filesize, printit, openit)
+    def ask_data_request(self, cb_answer, send_id, dtype: str, url: str, filesize: int,
+                         printit: bool, openit: bool) -> None:
+        GLib.idle_add(self.do_ask_data_request, cb_answer, send_id, dtype, url, filesize, printit, openit)
 
-    def do_ask_data_request(self, cb_answer, send_id, dtype: str, url: str, filesize: int, printit: bool, openit: bool):
+    def do_ask_data_request(self, cb_answer, send_id, dtype: str, url: str, filesize: int,
+                            printit: bool, openit: bool) -> None:
         from xpra.gtk.dialogs.open_requests import getOpenRequestsWindow
         timeout = self.remote_file_ask_timeout
 
@@ -527,23 +546,23 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
         self.file_ask_dialog.add_request(rec_answer, send_id, dtype, url, filesize, printit, openit, timeout)
         self.file_ask_dialog.show()
 
-    def close_ask_data_dialog(self):
+    def close_ask_data_dialog(self) -> None:
         fad = self.file_ask_dialog
         if fad:
             self.file_ask_dialog = None
             fad.close()
 
-    def show_ask_data_dialog(self, *_args):
+    def show_ask_data_dialog(self, *_args) -> None:
         from xpra.gtk.dialogs.open_requests import getOpenRequestsWindow
         self.file_ask_dialog = getOpenRequestsWindow(self.show_file_upload, self.cancel_download)
         self.file_ask_dialog.show()
 
-    def transfer_progress_update(self, send=True, transfer_id=0, elapsed=0, position=0, total=0, error=None):
+    def transfer_progress_update(self, send=True, transfer_id=0, elapsed=0, position=0, total=0, error=None) -> None:
         fad = self.file_ask_dialog
         if fad:
-            self.idle_add(fad.transfer_progress_update, send, transfer_id, elapsed, position, total, error)
+            GLib.idle_add(fad.transfer_progress_update, send, transfer_id, elapsed, position, total, error)
 
-    def accept_data(self, send_id, dtype: str, url: str, printit: bool, openit: bool):
+    def accept_data(self, send_id, dtype: str, url: str, printit: bool, openit: bool) -> tuple:
         # check if we have accepted this file via the GUI:
         r = self.data_send_requests.pop(send_id, None)
         if not r:
@@ -565,7 +584,7 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
         # return the printit and openit flag we got from the UI:
         return r[2], r[3]
 
-    def file_size_warning(self, action, location, basefilename, filesize, limit):
+    def file_size_warning(self, action: str, location: str, basefilename: str, filesize: int, limit: int) -> None:
         if self.file_size_dialog:
             # close previous warning
             self.file_size_dialog.close()
@@ -586,22 +605,22 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
         self.file_size_dialog.connect("response", self.close_file_size_warning)
         self.file_size_dialog.show()
 
-    def close_file_size_warning(self, *_args):
+    def close_file_size_warning(self, *_args) -> None:
         fsd = self.file_size_dialog
         if fsd:
             self.file_size_dialog = None
             fsd.close()
 
-    def download_server_log(self, callback: Callable = noop):
+    def download_server_log(self, callback: Callable = noop) -> None:
         filename = "${XPRA_SERVER_LOG}"
         self.file_request_callback[filename] = callback
         self.send_request_file(filename, self.open_files)
 
-    def send_download_request(self, *_args):
+    def send_download_request(self, *_args) -> None:
         command = ["xpra", "send-file"]
         self.send_start_command("Client-Download-File", command, True)
 
-    def show_file_upload(self, *args):
+    def show_file_upload(self, *args) -> None:
         if self.file_dialog:
             self.file_dialog.present()
             return
@@ -610,22 +629,25 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
         if self.remote_open_files:
             buttons += [Gtk.STOCK_OPEN, Gtk.ResponseType.ACCEPT]
         buttons += [Gtk.STOCK_OK, Gtk.ResponseType.OK]
-        self.file_dialog = Gtk.FileChooserDialog(
-            "File to upload",
-            parent=None,
-            action=Gtk.FileChooserAction.OPEN,
-            buttons=tuple(buttons))
-        self.file_dialog.set_default_response(Gtk.ResponseType.OK)
+        title = "File to upload"
+        if FILE_CHOOSER_NATIVE > 1 or (FILE_CHOOSER_NATIVE and not self.remote_open_files):
+            self.file_dialog = Gtk.FileChooserNative(title=title, action=Gtk.FileChooserAction.OPEN)
+            self.file_dialog.set_accept_label("Upload")
+        else:
+            self.file_dialog = Gtk.FileChooserDialog(title=title, action=Gtk.FileChooserAction.OPEN)
+            self.file_dialog.add_buttons(*buttons)
+            self.file_dialog.set_default_response(Gtk.ResponseType.OK)
         self.file_dialog.connect("response", self.file_upload_dialog_response)
         self.file_dialog.show()
 
-    def close_file_upload_dialog(self):
+    def close_file_upload_dialog(self) -> None:
         fd = self.file_dialog
-        if fd:
+        # the `FileChooserNative` is already closed automatically
+        if fd and hasattr(fd, "close"):
             fd.close()
             self.file_dialog = None
 
-    def file_upload_dialog_response(self, dialog, v):
+    def file_upload_dialog_response(self, dialog, v) -> None:
         if v not in (Gtk.ResponseType.OK, Gtk.ResponseType.ACCEPT):
             filelog(f"dialog response code {v}")
             self.close_file_upload_dialog()
@@ -665,8 +687,8 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
         about()
 
     def show_docs(self, *_args) -> None:
-        from xpra.scripts.main import run_docs
-        run_docs()
+        from xpra.scripts.main import show_docs
+        show_docs()
 
     def show_shortcuts(self, *_args) -> None:
         if self.shortcuts_info and not self.shortcuts_info.is_closed:
@@ -686,8 +708,7 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
             force_focus()
             self.session_info.present()
             return
-        p = self._protocol
-        conn = p._conn if p else None
+        conn = getattr(self._protocol, "._conn", None)
         from xpra.gtk.dialogs.session_info import SessionInfo
         self.session_info = SessionInfo(self, self.session_name, conn)
         self.session_info.set_args(*args)
@@ -703,7 +724,7 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
         from xpra.gtk.dialogs.bug_report import BugReport
         self.bug_report = BugReport()
 
-        def init_bug_report():
+        def init_bug_report() -> None:
             # skip things we aren't using:
             includes = {
                 "keyboard": bool(self.keyboard_helper),
@@ -728,9 +749,9 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
             self.bug_report.set_server_log_data(filedata)
 
         self.download_server_log(got_server_log)
-        self.timeout_add(200, init_bug_report)
+        GLib.timeout_add(200, init_bug_report)
 
-    def get_image(self, icon_name: str, size=None):
+    def get_image(self, icon_name: str, size=None) -> Gtk.Image | None:
         with log.trap_error(f"Error getting image for icon name {icon_name} and size {size}"):
             pixbuf = get_icon_pixbuf(icon_name)
             log(f"get_image({icon_name!r}, {size}) pixbuf={pixbuf}")
@@ -760,7 +781,7 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
         if not HAS_X11_BINDINGS:
             # nothing more we can do!
             return None
-        from xpra.x11.gtk_x11.prop import prop_get
+        from xpra.x11.gtk.prop import prop_get
         gdkwin = window.get_window()
         assert gdkwin
         v = prop_get(gdkwin.get_xid(), "_NET_FRAME_EXTENTS", ["u32"], ignore_errors=False)
@@ -831,13 +852,13 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
     def get_root_size(self):
         return get_root_size()
 
-    def get_raw_mouse_position(self):
+    def get_raw_mouse_position(self) -> tuple[int, int]:
         root = self.get_root_window()
         if not root:
             return -1, -1
         return root.get_pointer()[-3:-1]
 
-    def get_mouse_position(self):
+    def get_mouse_position(self) -> tuple[int, int]:
         p = self.get_raw_mouse_position()
         return self.cp(p[0], p[1])
 
@@ -925,12 +946,12 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
     def get_screen_sizes(self, xscale=1, yscale=1) -> list:
         return get_screen_sizes(xscale, yscale)
 
-    def reset_windows_cursors(self, *_args):
+    def reset_windows_cursors(self, *_args) -> None:
         cursorlog("reset_windows_cursors() resetting cursors for: %s", tuple(self._cursors.keys()))
         for w, cursor_data in tuple(self._cursors.items()):
             self.set_windows_cursor([w], cursor_data)
 
-    def set_windows_cursor(self, windows, cursor_data):
+    def set_windows_cursor(self, windows, cursor_data) -> None:
         cursorlog(f"set_windows_cursor({windows}, args[{len(cursor_data)}])")
         cursor = None
         if cursor_data:
@@ -956,7 +977,7 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
         if not display:
             return
         cursorlog("make_cursor(%s) has-name=%s, has-cursor-types=%s, xscale=%s, yscale=%s, USE_LOCAL_CURSORS=%s",
-                  ellipsizer(cursor_data),
+                  Ellipsizer(cursor_data),
                   len(cursor_data) >= 10, bool(cursor_types), self.xscale, self.yscale, USE_LOCAL_CURSORS)
         pixbuf = None
         if len(cursor_data) >= 10 and cursor_types:
@@ -1185,7 +1206,7 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
                             icon_name="opengl", callback=notify_callback)
 
         # wait for the main loop to run:
-        self.timeout_add(2 * 1000, delayed_notify)
+        GLib.timeout_add(2 * 1000, delayed_notify)
 
     # OpenGL bits:
     def init_opengl(self, enable_opengl: str) -> None:
@@ -1303,15 +1324,15 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
                 # log at warn level if the limit is low:
                 # (if we're likely to hit it - if the screen is as big or bigger)
                 w, h = self.get_root_size()
-                l = opengllog.info
+                log_fn = opengllog.info
                 if w * 2 <= self.gl_texture_size_limit and h * 2 <= self.gl_texture_size_limit:
-                    l = opengllog
+                    log_fn = opengllog.debug
                 if w >= self.gl_texture_size_limit or h >= self.gl_texture_size_limit:
-                    l = opengllog.warn
-                l("Warning: OpenGL windows will be clamped to the maximum texture size %ix%i",
-                  self.gl_texture_size_limit, self.gl_texture_size_limit)
+                    log_fn = opengllog.warn
+                log_fn("Warning: OpenGL windows will be clamped to the maximum texture size %ix%i",
+                       self.gl_texture_size_limit, self.gl_texture_size_limit)
                 glver = pver(self.opengl_props.get("opengl", ""))
-                l(f" for OpenGL {glver} renderer {renderer!r}")
+                log_fn(f" for OpenGL {glver} renderer {renderer!r}")
             if self.opengl_enabled and enable_opengl != "probe-success" and not self.opengl_force:
                 draw_result = test_gl_client_window(self.GLClientWindowClass,
                                                     max_window_size=self.max_window_size,
@@ -1342,7 +1363,7 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
             err("Error loading OpenGL support:", e)
 
     def get_client_window_classes(self, w: int, h: int, metadata: typedict,
-                                  override_redirect: bool) -> tuple[type, ...]:
+                                  override_redirect: bool) -> Sequence[type]:
         log("get_client_window_class%s", (w, h, metadata, override_redirect))
         log(" ClientWindowClass=%s, GLClientWindowClass=%s, opengl_enabled=%s, mmap_enabled=%s, encoding=%s",
             self.ClientWindowClass, self.GLClientWindowClass,
@@ -1428,29 +1449,15 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
         log(f"get_group_leader(..)={win}")
         if win:
             return win
-        pid = metadata.intget("pid", -1)
-        leader_xid = metadata.intget("group-leader-xid", -1)
-        log(f"get_group_leader: leader pid={pid}, xid={leader_xid}")
-        reftype = "xid"
-        ref : str | int = leader_xid
-        if ref < 0:
-            ci = metadata.strtupleget("class-instance")
-            if ci:
-                reftype = "class"
-                ref = "|".join(ci)
-            elif pid > 0:
-                reftype = "pid"
-                ref = pid
-            else:
-                # no reference to use
-                return None
-        refkey = f"{reftype}:{ref}"
+
+        refkey = get_group_ref(metadata)
+        log(f"get_group_leader: refkey={refkey}, metadata={metadata}, refs={self._ref_to_group_leader}")
         group_leader_window = self._ref_to_group_leader.get(refkey)
         if group_leader_window:
             log("found existing group leader window %s using ref=%s", group_leader_window, refkey)
             return group_leader_window
         # we need to create one:
-        title = "%s group leader for %s" % (self.session_name or "Xpra", pid)
+        title = "%s group leader for window %s" % (self.session_name or "Xpra", wid)
         # group_leader_window = Gdk.Window(None, 1, 1, Gtk.WindowType.TOPLEVEL, 0, Gdk.INPUT_ONLY, title)
         # static new(parent, attributes, attributes_mask)
         group_leader_window = GDKWindow(wclass=Gdk.WindowWindowClass.INPUT_ONLY, title=title)
@@ -1516,7 +1523,7 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
         cnt = self.clipboard_notification_timer
         if cnt:
             self.clipboard_notification_timer = 0
-            self.source_remove(cnt)
+            GLib.source_remove(cnt)
 
     def clipboard_notify(self, n: int) -> None:
         tray = self.tray
@@ -1536,9 +1543,9 @@ class GTKXpraClient(GObjectXpraClient, UIXpraClient):
             N = 1
             delay = max(0, round(1000 * (self.last_clipboard_notification + N - monotonic())))
 
-            self.clipboard_notification_timer = self.timeout_add(delay, self.reset_tray_icon)
+            self.clipboard_notification_timer = GLib.timeout_add(delay, self.reset_tray_icon)
 
-    def reset_tray_icon(self):
+    def reset_tray_icon(self) -> None:
         self.clipboard_notification_timer = 0
         tray = self.tray
         if not tray:
