@@ -16,6 +16,8 @@ Then it morphed into something quite messy. Sorry.
 import os.path
 import sys
 import traceback
+from typing import Any
+from collections.abc import Callable
 
 from xpra.scripts.config import read_config, make_defaults_struct, validate_config, save_config
 from xpra.gtk.signals import register_os_signals
@@ -28,8 +30,9 @@ from xpra.util.env import IgnoreWarningsContext
 from xpra.net.common import DEFAULT_PORT
 from xpra.util.thread import start_thread
 from xpra.gtk.dialogs.about import about
+from xpra.gtk.dialogs.util import hb_button
 from xpra.scripts.main import (
-    connect_to, make_client,
+    connect_to, make_client, bypass_no_gtk,
     configure_network, configure_env, configure_logging,
 )
 from xpra.common import noop
@@ -50,6 +53,7 @@ MODE_TCP = "tcp"
 MODE_SSL = "ssl"
 MODE_WS = "ws"
 MODE_WSS = "wss"
+MODE_QUIC = "quic"
 
 # what we save in the config file:
 SAVED_FIELDS = [
@@ -112,7 +116,7 @@ def set_history_from_active(optionmenu):
         optionmenu.set_history(i)
 
 
-def has_mdns():
+def has_mdns() -> bool:
     try:
         # pylint: disable=import-outside-toplevel
         from xpra.net.mdns import get_listener_class
@@ -131,6 +135,7 @@ def get_connection_modes():
         import ssl
         assert ssl
         modes.append(MODE_SSL)
+        modes.append(MODE_QUIC)
     except ImportError:
         pass
     # assume crypto is available
@@ -246,34 +251,36 @@ class ApplicationWindow:
         def show_about(*args):
             about(parent=self.window)
 
-        hb.add(button("About", "help-about", show_about))
+        hb.add(hb_button("About", "help-about", show_about))
         self.bug_tool = None
 
         def bug(*_args):
             if self.bug_tool is None:
-                from xpra.client.gtk3.bug_report import BugReport
+                from xpra.gtk.dialogs.bug_report import BugReport
                 self.bug_tool = BugReport()
                 self.bug_tool.init(show_about=False)
             self.bug_tool.show()
 
-        hb.add(button("Bug Report", "bugs", bug))
+        hb.add(hb_button("Bug Report", "bugs", bug))
         if has_mdns():
             self.mdns_gui = None
 
             def mdns(*_args):
                 if self.mdns_gui is None:
-                    from xpra.client.gtk3.mdns_gui import mdns_sessions
+                    from xpra.gtk.dialogs.mdns_gui import mdns_sessions
                     self.mdns_gui = mdns_sessions(self.config)
 
                     def close_mdns():
-                        self.mdns_gui.close()
-                        self.mdns_gui = None
+                        mdns_gui = self.mdns_gui
+                        if mdns_gui:
+                            self.mdns_gui = None
+                            mdns_gui.close()
 
                     self.mdns_gui.do_quit = close_mdns
                 else:
                     self.mdns_gui.present()
 
-            hb.add(button("Browse Sessions", "mdns", mdns))
+            hb.add(hb_button("Browse Sessions", "mdns", mdns))
         hb.show_all()
 
         vbox = Gtk.VBox(homogeneous=False, spacing=0)
@@ -625,7 +632,7 @@ class ApplicationWindow:
                 self.check_boxes_hbox.hide()
                 self.proxy_password_hbox.hide()
         self.validate()
-        if mode in (MODE_SSL, MODE_WSS) or (mode == MODE_SSH and not WIN32):
+        if mode in (MODE_SSL, MODE_WSS, MODE_QUIC) or (mode == MODE_SSH and not WIN32):
             self.nostrict_host_check.show()
         else:
             self.nostrict_host_check.hide()
@@ -775,13 +782,13 @@ class ApplicationWindow:
             params["display"] = f":{self.config.port}"
             params["display_name"] = f"unix-domain:{self.config.port}"
         else:
-            if self.config.mode not in (MODE_TCP, MODE_SSL, MODE_WS, MODE_WSS):
+            if self.config.mode not in (MODE_TCP, MODE_SSL, MODE_QUIC, MODE_WS, MODE_WSS):
                 raise ValueError(f"invalid / unsupported mode {self.config.mode}")
             params["host"] = self.config.host
             params["local"] = is_local(self.config.host)
             params["port"] = int(self.config.port)
             params["display_name"] = f"{self.config.mode}://{self.config.host}:{self.config.port}"
-            if self.config.mode in (MODE_SSL, MODE_WSS) and self.nostrict_host_check.get_active():
+            if self.config.mode in (MODE_SSL, MODE_WSS, MODE_QUIC) and self.nostrict_host_check.get_active():
                 params["strict-host-check"] = False
 
         # print("connect_to(%s)" % params)
@@ -798,6 +805,7 @@ class ApplicationWindow:
         def raise_exception(*args):
             raise RuntimeError(*args)
 
+        bypass_no_gtk()
         self.client = make_client(raise_exception, self.config)
         self.client.show_progress(30, "client configuration")
         self.client.init(self.config)
@@ -847,7 +855,6 @@ class ApplicationWindow:
         log("do_start_XpraClient(%s, %s) client=%s", conn, display_desc, self.client)
         self.client.encoding = self.config.encoding
         self.client.display_desc = display_desc
-        self.client.init_ui(self.config)
         self.client.setup_connection(conn)
         self.set_info_text("Network connection established")
         log("start_XpraClient() client initialized")
@@ -939,21 +946,23 @@ class ApplicationWindow:
         if self.exit_launcher:
             sys.exit(0)
 
-    def password_ok(self, *_args):
-        self.password_entry.modify_text(Gtk.StateType.NORMAL, black)
+    def password_ok(self, *_args) -> None:
+        with IgnoreWarningsContext():
+            self.password_entry.modify_text(Gtk.StateType.NORMAL, black)
 
-    def password_warning(self, *_args):
-        self.password_entry.modify_text(Gtk.StateType.NORMAL, red)
+    def password_warning(self, *_args) -> None:
+        with IgnoreWarningsContext():
+            self.password_entry.modify_text(Gtk.StateType.NORMAL, red)
         self.password_entry.grab_focus()
 
-    def set_widget_bg_color(self, widget, is_error=False):
+    def set_widget_bg_color(self, widget, is_error=False) -> None:
         with IgnoreWarningsContext():
             widget.modify_base(Gtk.StateType.NORMAL, red if is_error else white)
 
-    def set_widget_fg_color(self, widget, is_error=False):
+    def set_widget_fg_color(self, widget, is_error=False) -> None:
         modify_fg(widget, red if is_error else black)
 
-    def update_options_from_gui(self):
+    def update_options_from_gui(self) -> None:
         def pint(vstr):
             try:
                 return int(vstr)
@@ -979,7 +988,7 @@ class ApplicationWindow:
             self.config.mode = MODE_TCP
             if mode_enc.find("aes") > 0:
                 self.config.encryption = "AES-" + mode_enc.split("aes-")[1].upper()
-        elif mode_enc in (MODE_SSL, MODE_SSH, MODE_WS, MODE_WSS, MODE_NESTED_SSH):
+        elif mode_enc in (MODE_SSL, MODE_QUIC, MODE_SSH, MODE_WS, MODE_WSS, MODE_NESTED_SSH):
             self.config.mode = mode_enc
             self.config.encryption = ""
         log("update_options_from_gui() %s",
@@ -1027,13 +1036,13 @@ class ApplicationWindow:
         self.password_entry.set_text(password)
         self.host_entry.set_text(self.config.host)
 
-    def close_window(self, *_args):
+    def close_window(self, *_args) -> None:
         w = self.window
         if w:
             self.window = None
             w.destroy()
 
-    def destroy(self, *args):
+    def destroy(self, *args) -> bool:
         log("destroy%s", args)
         self.exit_launcher = True
         self.clean_client()
@@ -1041,11 +1050,11 @@ class ApplicationWindow:
         Gtk.main_quit()
         return False
 
-    def update_options_from_URL(self, url):
+    def update_options_from_URL(self, url: str) -> None:
         from xpra.scripts.parsing import parse_URL
         address, props = parse_URL(url)
         pa = address.split("://")
-        if pa[0] in (MODE_TCP, MODE_SSH, MODE_SSL, MODE_WS, MODE_WSS) and len(pa) >= 2:
+        if pa[0] in (MODE_TCP, MODE_SSH, MODE_SSL, MODE_QUIC, MODE_WS, MODE_WSS) and len(pa) >= 2:
             props["mode"] = pa[0]
             host = pa[1]
             if host.find("@") > 0:
@@ -1060,12 +1069,12 @@ class ApplicationWindow:
             props["host"] = host
         self._apply_props(props)
 
-    def update_options_from_file(self, filename):
+    def update_options_from_file(self, filename: str) -> None:
         log("update_options_from_file(%s)", filename)
         props = read_config(filename)
         self._apply_props(props)
 
-    def _apply_props(self, props):
+    def _apply_props(self, props: dict[str, Any]) -> None:
         # we rely on "ssh_port" being defined on the config object
         # so try to load it from file, and define it if not present:
         options = validate_config(props,
@@ -1078,16 +1087,16 @@ class ApplicationWindow:
         self.parse_ssh()
         log("_apply_props(%s) populated config with keys '%s', ssh=%s", props, options.keys(), self.config.ssh)
 
-    def choose_session_file(self, title, action, action_button, callback):
+    def choose_session_file(self, title: str, action, action_button, callback: Callable):
         file_filter = Gtk.FileFilter()
         file_filter.set_name("Xpra")
         file_filter.add_pattern("*.xpra")
         choose_file(self.window, title, action, action_button, callback, file_filter)
 
-    def save_clicked(self, *_args):
+    def save_clicked(self, *_args) -> None:
         self.update_options_from_gui()
 
-        def do_save(filename):
+        def do_save(filename) -> None:
             # make sure the file extension is .xpra
             if os.path.splitext(filename)[-1] != ".xpra":
                 filename += ".xpra"
@@ -1095,8 +1104,8 @@ class ApplicationWindow:
 
         self.choose_session_file("Save session settings to file", Gtk.FileChooserAction.SAVE, Gtk.STOCK_SAVE, do_save)
 
-    def load_clicked(self, *_args):
-        def do_load(filename):
+    def load_clicked(self, *_args) -> None:
+        def do_load(filename) -> None:
             self.update_options_from_file(filename)
             self.update_gui_from_config()
 
@@ -1105,7 +1114,7 @@ class ApplicationWindow:
 
 # on some platforms like win32, we don't have stdout
 # and this is a GUI application, so show a dialog with the error instead
-def exception_dialog(title):
+def exception_dialog(title: str) -> None:
     md = Gtk.MessageDialog(None, Gtk.DialogFlags.DESTROY_WITH_PARENT,
                            Gtk.MessageType.INFO, Gtk.ButtonsType.CLOSE, title)
     md.format_secondary_text(traceback.format_exc())
@@ -1120,7 +1129,7 @@ def exception_dialog(title):
     Gtk.main()
 
 
-def main(argv):
+def main(argv) -> int:
     from xpra.platform import program_context
     from xpra.log import enable_color
     with program_context("Xpra-Launcher", "Xpra Connection Launcher"):
@@ -1128,7 +1137,7 @@ def main(argv):
         return do_main(argv)
 
 
-def do_main(argv):
+def do_main(argv) -> int:
     from xpra.util.system import SIGNAMES
     from xpra.scripts.main import InitExit, InitInfo
     from xpra.platform.gui import init as gui_init, ready as gui_ready
