@@ -1,5 +1,5 @@
 # This file is part of Xpra.
-# Copyright (C) 2010-2023 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2024 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008, 2010 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -16,10 +16,10 @@ from dataclasses import dataclass
 from collections.abc import Callable
 
 from xpra.util.child_reaper import getChildReaper
-from xpra.os_util import POSIX, WIN32
+from xpra.os_util import POSIX, WIN32, gi_import
 from xpra.util.io import umask_context
 from xpra.util.objects import typedict
-from xpra.util.str_fn import csv, bytestostr
+from xpra.util.str_fn import csv
 from xpra.util.env import envint, envbool
 from xpra.scripts.config import str_to_bool, parse_with_unit
 from xpra.net.common import PacketType
@@ -27,6 +27,8 @@ from xpra.util.stats import std_unit
 from xpra.util.thread import start_thread
 from xpra.log import Logger
 
+GLib = gi_import("GLib")
+
 printlog = Logger("printing")
 filelog = Logger("file")
 
@@ -267,19 +269,13 @@ class FileTransferHandler(FileTransferAttributes):
         self.send_chunks_in_progress: dict[str, SendChunkState] = {}
         self.receive_chunks_in_progress: dict[str, ReceiveChunkState] = {}
         self.file_descriptors: set[int] = set()
-        if not getattr(self, "timeout_add", None):
-            from xpra.os_util import gi_import
-            glib = gi_import("GLib")
-            self.timeout_add = glib.timeout_add
-            self.idle_add = glib.idle_add
-            self.source_remove = glib.source_remove
 
     def cleanup(self) -> None:
         for t in self.pending_send_data_timers.values():
-            self.source_remove(t)
+            GLib.source_remove(t)
         self.pending_send_data_timers = {}
         for v in self.receive_chunks_in_progress.values():
-            self.source_remove(v.timer)
+            GLib.source_remove(v.timer)
         self.receive_chunks_in_progress = {}
         for x in tuple(self.file_descriptors):
             try:
@@ -362,7 +358,7 @@ class FileTransferHandler(FileTransferAttributes):
             timer = chunk_state.timer
             if timer:
                 chunk_state.timer = 0
-                self.source_remove(timer)
+                GLib.source_remove(timer)
             osclose(chunk_state.fd)
 
             # remove this transfer after a little while,
@@ -372,7 +368,7 @@ class FileTransferHandler(FileTransferAttributes):
                 self.receive_chunks_in_progress.pop(chunk_id, None)
                 return False
 
-            self.timeout_add(20000, clean_receive_state)
+            GLib.timeout_add(20000, clean_receive_state)
             filename = chunk_state.filename
             try:
                 os.unlink(filename)
@@ -388,7 +384,7 @@ class FileTransferHandler(FileTransferAttributes):
         file_data: bytes = packet[3]
         has_more = bool(packet[4])
         # if len(file_data)<1024:
-        #    from xpra.os_util import hexstr
+        #    from xpra.util.str_fn import hexstr
         #    filelog.warn("file_data=%s", hexstr(file_data))
         # filelog(f"file_data={len(file_data)} {type(file_data)}")
         filelog(f"file_data={len(file_data)} {type(file_data)}")
@@ -439,9 +435,9 @@ class FileTransferHandler(FileTransferAttributes):
         if has_more:
             progress(chunk_state.written)
             if chunk_state.timer:
-                self.source_remove(chunk_state.timer)
+                GLib.source_remove(chunk_state.timer)
             # remote end will send more after receiving the ack
-            chunk_state.timer = self.timeout_add(CHUNK_TIMEOUT, self._check_chunk_receiving, chunk_id, chunk)
+            chunk_state.timer = GLib.timeout_add(CHUNK_TIMEOUT, self._check_chunk_receiving, chunk_id, chunk)
             pct = round(100 * chunk_state.written / chunk_state.filesize)
             filelog(f"waiting for the next chunk, got {chunk_state.written:8} of {chunk_state.filesize:8}: {pct:3}%")
             return
@@ -551,12 +547,12 @@ class FileTransferHandler(FileTransferAttributes):
                 break
         if chunk_id:
             chunk = 0
-            l = len(self.receive_chunks_in_progress)
-            if l >= MAX_CONCURRENT_FILES:
-                self.cancel_file(chunk_id, f"too many file transfers in progress: {l}", chunk)
+            nfiles = len(self.receive_chunks_in_progress)
+            if nfiles >= MAX_CONCURRENT_FILES:
+                self.cancel_file(chunk_id, f"too many file transfers in progress: {nfiles}", chunk)
                 osclose(fd)
                 return
-            timer = self.timeout_add(CHUNK_TIMEOUT, self._check_chunk_receiving, chunk_id, chunk)
+            timer = GLib.timeout_add(CHUNK_TIMEOUT, self._check_chunk_receiving, chunk_id, chunk)
             self.receive_chunks_in_progress[chunk_id] = ReceiveChunkState(monotonic(),
                                                                           fd, filename, mimetype,
                                                                           printit, openit, filesize,
@@ -596,7 +592,7 @@ class FileTransferHandler(FileTransferAttributes):
         rf = options.tupleget("request-file")
         if rf and len(rf) >= 2:
             argf = rf[0]
-            cb = self.file_request_callback.pop(bytestostr(argf), None)
+            cb = self.file_request_callback.pop(argf, None)
             if cb:
                 cb(filename, filesize)
                 return
@@ -679,7 +675,7 @@ class FileTransferHandler(FileTransferAttributes):
 
         if check_printing_finished():
             # check every 10 seconds:
-            self.timeout_add(10000, check_printing_finished)
+            GLib.timeout_add(10000, check_printing_finished)
 
     def _open_file(self, url: str) -> None:
         filelog("_open_file(%s)", url)
@@ -771,13 +767,13 @@ class FileTransferHandler(FileTransferAttributes):
     def send_file(self, filename, mimetype, data, filesize=0,
                   printit=False, openit=False, options=None):
         if printit:
-            l = printlog
             if not self.printing:
-                l.warn("Warning: printing is not enabled for %s", self)
+                printlog.warn("Warning: printing is not enabled for %s", self)
                 return False
             if not self.remote_printing:
-                l.warn("Warning: remote end does not support printing")
+                printlog.warn("Warning: remote end does not support printing")
                 return False
+            logger = printlog
             ask = self.remote_printing_ask
             action = "print"
         else:
@@ -785,15 +781,15 @@ class FileTransferHandler(FileTransferAttributes):
                 filelog.warn("Warning: file transfers are not enabled for %s", self)
                 return False
             if not self.remote_file_transfer:
-                printlog.warn("Warning: remote end does not support file transfers")
+                filelog.warn("Warning: remote end does not support file transfers")
                 return False
-            l = filelog
+            logger = filelog
             ask = self.remote_file_transfer_ask
             action = "upload"
             if openit:
                 if not self.remote_open_files:
-                    l.info("opening the file after transfer is disabled on the remote end")
-                    l.info(" sending only, the file will need to be opened manually")
+                    logger.info("opening the file after transfer is disabled on the remote end")
+                    logger.info(" sending only, the file will need to be opened manually")
                     openit = False
                     action = "upload"
                 else:
@@ -801,8 +797,8 @@ class FileTransferHandler(FileTransferAttributes):
                     action = "open"
         assert len(data) >= filesize, "data is smaller then the given file size!"
         data = data[:filesize]  # gio may null terminate it
-        l("send_file%s action=%s, ask=%s",
-          (filename, mimetype, type(data), f"{filesize} bytes", printit, openit, options), action, ask)
+        logger("send_file%s action=%s, ask=%s",
+               (filename, mimetype, type(data), f"{filesize} bytes", printit, openit, options), action, ask)
         self.dump_remote_caps()
         if not self.check_file_size(action, filename, filesize):
             return False
@@ -822,7 +818,7 @@ class FileTransferHandler(FileTransferAttributes):
         spd = SendPendingData(dtype, url, mimetype, data, filesize, printit, openit, options or {})
         self.pending_send_data[send_id] = spd
         delay = self.remote_file_ask_timeout * 1000
-        self.pending_send_data_timers[send_id] = self.timeout_add(delay, self.send_data_ask_timeout, send_id)
+        self.pending_send_data_timers[send_id] = GLib.timeout_add(delay, self.send_data_ask_timeout, send_id)
         filelog(f"sending data request for {dtype} {url!r} with send-id {send_id!r}")
         self.send("send-data-request", dtype, send_id, url, mimetype, filesize, printit, openit, options or {})
         return send_id
@@ -847,7 +843,7 @@ class FileTransferHandler(FileTransferAttributes):
         rf = options.tupleget("request-file")
         if rf and len(rf) >= 2:
             argf, openit = rf[:2]
-            openit = self.files_requested.pop(bytestostr(argf), None)
+            openit = self.files_requested.pop(argf, None)
             if openit is not None:
                 self.files_accepted[send_id] = openit
                 cb_answer(True)
@@ -894,7 +890,7 @@ class FileTransferHandler(FileTransferAttributes):
         filelog("process send-data-response: send_id=%s, accept=%s", send_id, accept)
         timer = self.pending_send_data_timers.pop(send_id, None)
         if timer:
-            self.source_remove(timer)
+            GLib.source_remove(timer)
         try:
             spd: SendPendingData = self.pending_send_data[send_id]
         except KeyError:
@@ -937,11 +933,11 @@ class FileTransferHandler(FileTransferAttributes):
                      printit: bool = False, openit: bool = False, options=None, send_id: str = "") -> bool:
         if printit:
             action = "print"
-            l = printlog
+            logger = printlog
         else:
             action = "upload"
-            l = filelog
-        l("do_send_file%s", (filename, mimetype, type(data), f"{filesize} bytes", printit, openit, options))
+            logger = filelog
+        logger("do_send_file%s", (filename, mimetype, type(data), f"{filesize} bytes", printit, openit, options))
         if not self.check_file_size(action, filename, filesize):
             return False
         h = hashlib.sha256()
@@ -960,7 +956,7 @@ class FileTransferHandler(FileTransferAttributes):
             options["file-chunk-id"] = chunk_id
             # timer to check that the other end is requesting more chunks:
             chunk_no = 0
-            timer = self.timeout_add(CHUNK_TIMEOUT, self._check_chunk_sending, chunk_id, chunk_no)
+            timer = GLib.timeout_add(CHUNK_TIMEOUT, self._check_chunk_sending, chunk_id, chunk_no)
             self.send_chunks_in_progress[chunk_id] = SendChunkState(monotonic(), data, chunk_size, timer, chunk_no)
             cdata = b""
             filelog("using chunks, sending initial file-chunk-id=%s, for chunk size=%s",
@@ -994,7 +990,7 @@ class FileTransferHandler(FileTransferAttributes):
         timer = chunk_state.timer
         if timer:
             chunk_state.timer = 0
-            self.source_remove(timer)
+            GLib.source_remove(timer)
 
     def _process_ack_file_chunk(self, packet: PacketType) -> None:
         # the other end received our send-file or send-file-chunk,
@@ -1032,8 +1028,8 @@ class FileTransferHandler(FileTransferAttributes):
         chunk += 1
         chunk_state.chunk = chunk
         if chunk_state.timer:
-            self.source_remove(chunk_state.timer)
-        chunk_state.timer = self.timeout_add(CHUNK_TIMEOUT, self._check_chunk_sending, chunk_id, chunk)
+            GLib.source_remove(chunk_state.timer)
+        chunk_state.timer = GLib.timeout_add(CHUNK_TIMEOUT, self._check_chunk_sending, chunk_id, chunk)
         self.send("send-file-chunk", chunk_id, chunk, cdata, bool(chunk_state.data))
 
     def send(self, *parts) -> None:
