@@ -1,18 +1,21 @@
 # This file is part of Xpra.
-# Copyright (C) 2016-2023 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2016-2024 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import os
+import sys
 import glob
 import posixpath
 import mimetypes
+import socket
 from urllib.parse import unquote
 from http.server import BaseHTTPRequestHandler
 from typing import Any
-from collections.abc import Iterable
+from collections.abc import Iterable, Callable
 
-from xpra.common import DEFAULT_XDG_DATA_DIRS
+from xpra.common import DEFAULT_XDG_DATA_DIRS, FULL_INFO
+from xpra.net.common import HttpResponse
 from xpra.net.http.common import EXTENSION_TO_MIMETYPE
 from xpra.net.http.directory_listing import list_directory
 from xpra.net.bytestreams import pretty_socket
@@ -32,28 +35,11 @@ AUTH_USERNAME = os.environ.get("XPRA_HTTP_AUTH_USERNAME", "")
 AUTH_PASSWORD = os.environ.get("XPRA_HTTP_AUTH_PASSWORD", "")
 
 
-# should be converted to use standard library
-def parse_url(handler) -> dict[str, str]:
-    try:
-        args_str = handler.path.split("?", 1)[1]
-    except IndexError:
-        return {}
-    # parse args:
-    args = {}
-    for x in args_str.split("&"):
-        v = x.split("=", 1)
-        if len(v) == 1:
-            args[v[0]] = ""
-        else:
-            args[v[0]] = v[1]
-    return args
-
-
 http_headers_cache: dict[str, str] = {}
 http_headers_time: dict[str, float] = {}
 
 
-def may_reload_headers(http_headers_dirs):
+def may_reload_headers(http_headers_dirs: Iterable[str]) -> dict[str, str]:
     mtimes: dict[str, float] = {}
     global http_headers_cache
     if http_headers_cache:
@@ -224,11 +210,14 @@ class HTTPRequestHandler(BaseHTTPRequestHandler):
 
     wbufsize = 0  # we flush explicitly when needed
     server_version = "Xpra-HTTP-Server"
+    sys_version = "Python/" + ".".join(str(vnum) for vnum in sys.version_info[:FULL_INFO+1])
 
     def __init__(self, sock, addr,
                  web_root: str = "/usr/share/xpra/www/",
-                 http_headers_dirs: Iterable[str] = ("/etc/xpra/http-headers",), script_paths=None,
-                 username: str = AUTH_USERNAME, password: str = AUTH_PASSWORD):
+                 http_headers_dirs: Iterable[str] = ("/etc/xpra/http-headers", ),
+                 script_paths: dict[str, Callable[[str], HttpResponse]] = None,
+                 username: str = AUTH_USERNAME,
+                 password: str = AUTH_PASSWORD):
         self.web_root = web_root
         self.http_headers_dirs = http_headers_dirs
         self.script_paths = script_paths or {}
@@ -242,7 +231,9 @@ class HTTPRequestHandler(BaseHTTPRequestHandler):
 
     def log_error(self, fmt, *args) -> None:  # pylint: disable=arguments-differ
         # don't log 404s at error level:
-        if len(args) == 2 and args[0] == 404:
+        if len(args) == 1 and isinstance(args[0], TimeoutError):
+            log(fmt, *args)
+        elif len(args) == 2 and args[0] == 404:
             log(fmt, *args)
         else:
             log.error(fmt, *args)
@@ -271,7 +262,7 @@ class HTTPRequestHandler(BaseHTTPRequestHandler):
                 self.send_header(k, v)
         super().end_headers()
 
-    def get_headers(self):
+    def get_headers(self) -> dict[str, str]:
         return may_reload_headers(self.http_headers_dirs)
 
     def do_POST(self) -> None:
@@ -311,7 +302,7 @@ class HTTPRequestHandler(BaseHTTPRequestHandler):
         import base64
         try:
             s = base64.b64decode(b64str).decode("utf8")
-        except Exception:
+        except ValueError:
             s = ""
         if s.find(":") < 0:
             return auth_err("invalid authentication format")
@@ -330,7 +321,8 @@ class HTTPRequestHandler(BaseHTTPRequestHandler):
         if content:
             try:
                 self.wfile.write(content)
-            except (BrokenPipeError, ConnectionResetError) as e:
+            except (BrokenPipeError, ConnectionResetError, socket.error) as e:
+                # ssl.SSLEOFError is a socket.error
                 log("handle_request() %s", e)
             except TypeError:
                 self.close_connection = True
@@ -360,7 +352,7 @@ class HTTPRequestHandler(BaseHTTPRequestHandler):
             log(f"failed to send {code} error - maybe some of the headers were already sent?", exc_info=True)
 
     # code taken from MIT licensed code in GzipSimpleHTTPServer.py
-    def send_head(self):
+    def send_head(self) -> bytes:
         path = self.path.split("?", 1)[0].split("#", 1)[0]
         # strip path after second slash:
         script_path = path
@@ -375,7 +367,7 @@ class HTTPRequestHandler(BaseHTTPRequestHandler):
             except Exception:
                 log.error(f"Error calling script {script}", exc_info=True)
                 self.send_error(500, "Server error")
-                return None
+                return b""
             self.send_response(code)
             self.extra_headers.update(headers or {})
             self.end_headers()
@@ -383,14 +375,14 @@ class HTTPRequestHandler(BaseHTTPRequestHandler):
         path = translate_path(self.path, self.web_root)
         if not path or not os.path.exists(path):
             self.send_error(404, "Path not found")
-            return None
+            return b""
         if os.path.isdir(path):
             if not path.endswith('/'):
                 # redirect browser - doing basically what apache does
                 self.send_response(301)
                 self.send_header("Location", path + "/")
                 self.end_headers()
-                return None
+                return b""
             for index in "index.html", "index.htm":
                 index = os.path.join(path, index)
                 if os.path.exists(index):
@@ -399,9 +391,12 @@ class HTTPRequestHandler(BaseHTTPRequestHandler):
             else:
                 if not self.directory_listing:
                     self.send_error(403, "Directory listing forbidden")
-                    return None
-                return list_directory(path).read()
-
+                    return b""
+                code, headers, body = list_directory(path)
+                self.send_response(code)
+                self.extra_headers.update(headers)
+                self.end_headers()
+                return body
         try:
             accept_encoding = self.headers.get("accept-encoding", "").split(",")
             code, extra_headers, content = load_path(accept_encoding, path)
@@ -411,6 +406,7 @@ class HTTPRequestHandler(BaseHTTPRequestHandler):
             self.send_response(code)
             self.extra_headers.update(extra_headers)
             self.end_headers()
+            return content
         except OSError as e:
             self.close_connection = True
             log("send_head()", exc_info=True)
@@ -421,5 +417,4 @@ class HTTPRequestHandler(BaseHTTPRequestHandler):
             else:
                 log.estr(e)
             self.send_error(404, "File not found")
-            return None
-        return content
+            return b""
