@@ -9,16 +9,16 @@ from time import sleep, monotonic
 from ctypes import Structure, c_uint8, sizeof
 from typing import Any
 from importlib.util import find_spec
-from collections.abc import Callable, ByteString
+from collections.abc import Callable
 
-from xpra.common import GROUP, SocketState, noerr
+from xpra.common import GROUP, SocketState, noerr, SizedBuffer
 from xpra.scripts.config import InitException, InitExit, TRUE_OPTIONS
 from xpra.exit_codes import ExitCode
 from xpra.net.common import DEFAULT_PORT, AUTO_ABSTRACT_SOCKET, ABSTRACT_SOCKET_PREFIX
 from xpra.net.bytestreams import set_socket_timeout, pretty_socket, SocketConnection, SOCKET_TIMEOUT
 from xpra.os_util import getuid, get_username_for_uid, get_groups, get_group_id, gi_import, WIN32, OSX, POSIX
 from xpra.util.io import path_permission_info, umask_context
-from xpra.util.str_fn import csv
+from xpra.util.str_fn import csv, memoryview_to_bytes
 from xpra.util.parsing import parse_simple_dict
 from xpra.util.env import envint, envbool, SilenceWarningsContext
 from xpra.util.thread import start_thread
@@ -149,7 +149,7 @@ def hosts(host_str: str) -> list[str]:
     return [host_str]
 
 
-def add_listen_socket(socktype: str, sock, info, server, new_connection_cb, options=None) -> Callable | None:
+def add_listen_socket(socktype: str, sock, info, server, new_connection_cb: Callable, options: dict) -> Callable | None:
     log = get_network_logger()
     log("add_listen_socket%s", (socktype, sock, info, server, new_connection_cb, options))
     try:
@@ -182,7 +182,7 @@ def add_listen_socket(socktype: str, sock, info, server, new_connection_cb, opti
                 from xpra.net.upnp import upnp_add
                 upnp_cleanup.append(upnp_add(socktype, info, options))
 
-        def cleanup():
+        def cleanup() -> None:
             for source in tuple(sources):
                 GLib.source_remove(source)
                 sources.remove(source)
@@ -291,7 +291,7 @@ def get_sockopt_tcp_info(sock, sockopt_tcpinfo: int, attributes=POSIX_TCP_INFO)
     return d
 
 
-def looks_like_xpra_packet(data: ByteString) -> bool:
+def looks_like_xpra_packet(data: bytes) -> bool:
     if len(data) < 8:
         return False
     if data[0] != ord("P"):
@@ -329,9 +329,10 @@ def looks_like_xpra_packet(data: ByteString) -> bool:
     return True
 
 
-def guess_packet_type(data: ByteString) -> str:
-    if not data:
+def guess_packet_type(buf: SizedBuffer) -> str:
+    if not buf:
         return ""
+    data = memoryview_to_bytes(buf[:32])
     if looks_like_xpra_packet(data):
         return "xpra"
     if data[:4] == b"SSH-":
@@ -348,7 +349,8 @@ def guess_packet_type(data: ByteString) -> str:
     return ""
 
 
-def create_sockets(opts, error_cb: Callable, retry: int = 0) -> dict[Any, dict]:
+def create_sockets(opts, error_cb: Callable, retry: int = 0,
+                   sd_listen=POSIX and not OSX, ssh_upgrades=True) -> dict[Any, dict]:
     bind_tcp = parse_bind_ip(opts.bind_tcp)
     bind_ssl = parse_bind_ip(opts.bind_ssl, 443)
     bind_ssh = parse_bind_ip(opts.bind_ssh, 22)
@@ -362,7 +364,6 @@ def create_sockets(opts, error_cb: Callable, retry: int = 0) -> dict[Any, dict]:
     # Initialize the TCP sockets before the display,
     # That way, errors won't make us kill the Xvfb
     # (which may not be ours to kill at that point)
-    ssh_upgrades = opts.ssh_upgrade
     if ssh_upgrades:
         err = ""
         try:
@@ -375,8 +376,8 @@ def create_sockets(opts, error_cb: Callable, retry: int = 0) -> dict[Any, dict]:
             from xpra.log import Logger
             sshlog = Logger("ssh")
             sshlog("import paramiko", exc_info=True)
-            sshlog.error("Error: cannot enable SSH socket upgrades")
-            sshlog.estr(err)
+            sshlog.warn("Error: cannot enable SSH socket upgrades")
+            sshlog.warn(" %s", err)
             opts.ssh_upgrade = False
     log = get_network_logger()
     # prepare tcp socket definitions:
@@ -431,7 +432,7 @@ def create_sockets(opts, error_cb: Callable, retry: int = 0) -> dict[Any, dict]:
         sockets[sock] = options
 
     # systemd socket activation:
-    if POSIX and not OSX:
+    if sd_listen:
         try:
             from xpra.platform.posix.sd_listen import get_sd_listen_sockets
         except ImportError as e:
@@ -550,7 +551,9 @@ def parse_bind_ip(bind_ip: list[str], default_port: int = DEFAULT_PORT) -> dict[
     ip_sockets: dict[tuple[str, int], dict] = {}
     if bind_ip:
         for spec in bind_ip:
+            # ie: "127.0.0.1:10000,someoption=somevalue"
             parts = spec.split(",", 1)
+            # ie: "127.0.0.1:10000"
             ip_port = parts[0]
             if ":" not in spec:
                 raise InitException("port must be specified as [HOST]:PORT")
@@ -739,7 +742,7 @@ def setup_local_sockets(bind, socket_dir: str, socket_dirs, session_dir: str,
                 defs[("named-pipe", npl, sockpath, npl.stop)] = options
         else:
 
-            def checkstate(sockpath: str, state: SocketState | str):
+            def checkstate(sockpath: str, state: SocketState | str) -> None:
                 if state not in (SocketState.DEAD, SocketState.UNKNOWN):
                     if state == SocketState.INACCESSIBLE:
                         raise InitException(f"An xpra server is already running at {sockpath!r}\n")
@@ -786,7 +789,7 @@ def setup_local_sockets(bind, socket_dir: str, socket_dirs, session_dir: str,
                 # so we can do them in parallel:
                 threads = []
 
-                def timeout_probe(sockpath: str):
+                def timeout_probe(sockpath: str) -> None:
                     # we need a loop because "DEAD" sockets may return immediately
                     # (ie: when the server is starting up)
                     start = monotonic()
@@ -866,7 +869,7 @@ def setup_local_sockets(bind, socket_dir: str, socket_dirs, session_dir: str,
                 sock_str = f"{sock_def[0]} {sock_def[2]!r}"
                 cleanup_socket = sock_def[-1]
                 cleanup_socket()
-            except Exception:
+            except (IndexError, ValueError, OSError):
                 log(f"error cleaning up socket {sock_str}", exc_info=True)
                 log.error(f"Error cleaning up socket {sock_str}:", exc_info=True)
                 log.error(f" using {sock_def}")
@@ -909,12 +912,12 @@ def handle_socket_error(sockpath: str, sperms: int, e) -> None:
             raise InitExit(ExitCode.SOCKET_CREATION_ERROR, f"failed to create socket {sockpath}")
 
 
-def socket_connect(host: str, port: int, timeout: float = SOCKET_TIMEOUT):
+def socket_connect(host: str, port: int, timeout: float = SOCKET_TIMEOUT) -> socket.socket | None:
     socktype = socket.SOCK_STREAM
     family = 0  # 0 means any
     try:
         addrinfo = socket.getaddrinfo(host, port, family, socktype)
-    except Exception as e:
+    except OSError as e:
         stypestr = {
             socket.AF_INET6: "IPv6",
             socket.AF_INET: "IPv4",
@@ -934,7 +937,7 @@ def socket_connect(host: str, port: int, timeout: float = SOCKET_TIMEOUT):
             sock.connect(sockaddr)
             sock.settimeout(None)
             return sock
-        except Exception:
+        except OSError:
             log("failed to connect using %s%s for %s", sock.connect, sockaddr, addr, exc_info=True)
             noerr(sock.close)
     return None
