@@ -7,13 +7,16 @@
 import os
 import sys
 from typing import Any
-from collections.abc import Callable
+from collections.abc import Callable, Sequence
 
+from xpra.os_util import gi_import
 from xpra.util.system import is_Wayland, is_X11
 from xpra.util.str_fn import csv, bytestostr
 from xpra.util.env import envint, envbool, first_time, get_saved_env, get_saved_env_var
 from xpra.log import Logger
 
+GLib = gi_import("GLib")
+
 log = Logger("posix")
 eventlog = Logger("posix", "events")
 screenlog = Logger("posix", "screen")
@@ -103,7 +106,7 @@ def get_clipboard_native_class() -> str:
         assert x11
     except ImportError:
         return gtk_clipboard_class
-    return "xpra.x11.gtk_x11.clipboard.X11Clipboard"
+    return "xpra.x11.gtk.clipboard.X11Clipboard"
 
 
 def get_native_system_tray_classes() -> list[type]:
@@ -168,8 +171,11 @@ def _get_xsettings():
     return None
 
 
-def _get_xsettings_dict():
-    from xpra.x11.common import xsettings_to_dict
+def _get_xsettings_dict() -> dict[str, Any]:
+    try:
+        from xpra.x11.common import xsettings_to_dict
+    except ImportError:
+        return {}
     return xsettings_to_dict(_get_xsettings())
 
 
@@ -298,13 +304,13 @@ def get_number_of_desktops() -> int:
     return 0
 
 
-def get_desktop_names() -> tuple[str, ...]:
+def get_desktop_names() -> Sequence[str]:
     if x11_bindings():
         from xpra.x11.common import get_desktop_names as get_x11_desktop_names
         from xpra.gtk.error import xsync
         with xsync:
             return get_x11_desktop_names()
-    return ("Main",)
+    return ("Main", )
 
 
 def get_vrefresh() -> int:
@@ -435,8 +441,8 @@ WINDOW_ADD_HOOKS: list[Callable] = []
 
 
 def add_window_hooks(window) -> None:
-    for x in WINDOW_ADD_HOOKS:
-        x(window)
+    for callback in WINDOW_ADD_HOOKS:
+        callback(window)
     log("add_window_hooks(%s) added %s", window, WINDOW_ADD_HOOKS)
 
 
@@ -444,8 +450,8 @@ WINDOW_REMOVE_HOOKS: list[Callable] = []
 
 
 def remove_window_hooks(window):
-    for x in WINDOW_REMOVE_HOOKS:
-        x(window)
+    for callback in WINDOW_REMOVE_HOOKS:
+        callback(window)
     log("remove_window_hooks(%s) added %s", window, WINDOW_REMOVE_HOOKS)
 
 
@@ -477,7 +483,7 @@ class XI2_Window:
         self.X11Window = X11WindowBindings()
         self.window = window
         self.xid = window.get_window().get_xid()
-        self.windows: tuple[int, ...] = ()
+        self.windows: Sequence[int] = ()
         self.motion_valuators = {}
         window.connect("configure-event", self.configured)
         self.configured()
@@ -512,7 +518,7 @@ class XI2_Window:
     def do_xi_hierarchy_changed(self, *_args) -> None:
         self.motion_valuators = {}
 
-    def get_parent_windows(self, oxid: int) -> tuple[int, ...]:
+    def get_parent_windows(self, oxid: int) -> Sequence[int]:
         windows = [oxid]
         root = self.X11Window.get_root_xid()
         xid = oxid
@@ -664,7 +670,7 @@ class ClientExtras:
         if self.x11_filter:
             return
         try:
-            from xpra.x11.gtk3.bindings import init_x11_filter  # @UnresolvedImport, @UnusedImport
+            from xpra.x11.gtk.bindings import init_x11_filter  # @UnresolvedImport, @UnusedImport
             self.x11_filter = init_x11_filter()
             log("x11_filter=%s", self.x11_filter)
         except Exception as e:
@@ -677,7 +683,7 @@ class ClientExtras:
         log("cleanup() xsettings_watcher=%s, root_props_watcher=%s", self._xsettings_watcher, self._root_props_watcher)
         if self.x11_filter:
             self.x11_filter = None
-            from xpra.x11.gtk3.bindings import cleanup_x11_filter  # @UnresolvedImport, @UnusedImport
+            from xpra.x11.gtk.bindings import cleanup_x11_filter  # @UnresolvedImport, @UnusedImport
             cleanup_x11_filter()
         if self._xsettings_watcher:
             self._xsettings_watcher.cleanup()
@@ -843,7 +849,7 @@ class ClientExtras:
             self.client.send_input_devices("xi", devices)
 
     def setup_xi(self) -> None:
-        self.client.timeout_add(100, self.do_setup_xi)
+        GLib.timeout_add(100, self.do_setup_xi)
 
     def do_setup_xi(self) -> bool:
         if self.client.server_input_devices not in ("xi", "uinput"):
@@ -903,7 +909,7 @@ class ClientExtras:
     def get_resource_manager(self):
         try:
             from xpra.gtk.util import get_default_root_window
-            from xpra.x11.gtk_x11.prop import prop_get
+            from xpra.x11.gtk.prop import prop_get
             root = get_default_root_window()
             xid = root.get_xid()
             value = prop_get(xid, "RESOURCE_MANAGER", "latin1", ignore_errors=True)
@@ -931,7 +937,7 @@ class ClientExtras:
 
 def main() -> int:
     try:
-        from xpra.x11.gtk3.display_source import init_gdk_display_source
+        from xpra.x11.gtk.display_source import init_gdk_display_source
         init_gdk_display_source()
     except ImportError:
         pass
