@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 # This file is part of Xpra.
-# Copyright (C) 2010-2023 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2024 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -10,7 +10,7 @@ import os
 import glob
 import shlex
 from typing import Any
-from collections.abc import Callable, Iterable
+from collections.abc import Callable, Iterable, Sequence
 
 from xpra.common import noop
 from xpra.util.str_fn import csv
@@ -49,10 +49,10 @@ class InitExit(Exception):
 
 DEBUG_CONFIG_PROPERTIES: list[str] = os.environ.get("XPRA_DEBUG_CONFIG_PROPERTIES", "").split()
 
-DEFAULT_XPRA_CONF_FILENAME : str = os.environ.get("XPRA_CONF_FILENAME", 'xpra.conf')
-DEFAULT_NET_WM_NAME : str = os.environ.get("XPRA_NET_WM_NAME", "Xpra")
+DEFAULT_XPRA_CONF_FILENAME: str = os.environ.get("XPRA_CONF_FILENAME", 'xpra.conf')
+DEFAULT_NET_WM_NAME: str = os.environ.get("XPRA_NET_WM_NAME", "Xpra")
 
-DEFAULT_POSTSCRIPT_PRINTER : str = ""
+DEFAULT_POSTSCRIPT_PRINTER: str = ""
 if POSIX:
     DEFAULT_POSTSCRIPT_PRINTER = os.environ.get("XPRA_POSTSCRIPT_PRINTER", "drv:///sample.drv/generic.ppd")
 DEFAULT_PULSEAUDIO = None   # auto
@@ -63,15 +63,15 @@ if OSX or WIN32:   # pragma: no cover
 
 
 def remove_dupes(seq: Iterable[Any]) -> list[Any]:
-    seen : set[Any] = set()
-    seen_add : Callable = seen.add
+    seen: set[Any] = set()
+    seen_add: Callable = seen.add
     return [x for x in seq if not (x in seen or seen_add(x))]
 
 
-_has_audio_support = None
+_has_audio_support: bool | None = None
 
 
-def has_audio_support():
+def has_audio_support() -> bool:
     global _has_audio_support
     if _has_audio_support is None:
         try:
@@ -79,7 +79,7 @@ def has_audio_support():
             _has_audio_support = bool(xpra.audio)
         except ImportError:
             _has_audio_support = False
-    return _has_audio_support
+    return bool(_has_audio_support)
 
 
 def find_html5_path() -> str:
@@ -143,7 +143,7 @@ def get_xorg_bin() -> str:
     return ""
 
 
-def get_Xdummy_confdir():
+def get_Xdummy_confdir() -> str:
     from xpra.platform.posix.paths import get_runtime_dir
     xrd = get_runtime_dir()
     if xrd:
@@ -178,6 +178,7 @@ def get_Xvfb_command(width=8192, height=4096, dpi=96) -> list[str]:
         "Xvfb",
         "+extension", "GLX",
         "+extension", "Composite",
+        "+extension", "RANDR", "+extension", "RENDER",
         "-screen", "0", f"{width}x{height}x24+32",
         # better than leaving to vfb after a resize?
         "-nolisten", "tcp",
@@ -190,8 +191,8 @@ def get_Xvfb_command(width=8192, height=4096, dpi=96) -> list[str]:
 
 
 def detect_xvfb_command(conf_dir="/etc/xpra/", bin_dir="",
-                        Xdummy_ENABLED: bool | None = None, Xdummy_wrapper_ENABLED: bool | None=None,
-                        warn_fn:Callable = warn) -> list[str]:
+                        Xdummy_ENABLED: bool | None = None, Xdummy_wrapper_ENABLED: bool | None = None,
+                        warn_fn: Callable = warn) -> list[str]:
     """
     This function returns the xvfb command to use.
     It can either be an `Xvfb` command or one that uses `Xdummy`,
@@ -201,7 +202,7 @@ def detect_xvfb_command(conf_dir="/etc/xpra/", bin_dir="",
         return []
     if OSX:     # pragma: no cover
         return get_Xvfb_command()
-    if sys.platform.find("bsd")>=0 and Xdummy_ENABLED is None:  # pragma: no cover
+    if sys.platform.find("bsd") >= 0 and Xdummy_ENABLED is None:  # pragma: no cover
         warn_fn(f"Warning: sorry, no support for Xdummy on {sys.platform}")
         return get_Xvfb_command()
     # if is_arm():
@@ -238,8 +239,8 @@ def detect_xdummy_command(conf_dir="/etc/xpra/", bin_dir="",
         # auto-detect
         import stat
         xorg_stat = os.stat(xorg_bin)
-        if (xorg_stat.st_mode & stat.S_ISUID)!=0:
-            if (xorg_stat.st_mode & stat.S_IROTH)==0:
+        if (xorg_stat.st_mode & stat.S_ISUID) != 0:
+            if (xorg_stat.st_mode & stat.S_IROTH) == 0:
                 warn_fn(f"{xorg_bin} is suid and not readable, Xdummy support unavailable")
                 return get_Xvfb_command()
             debug(f"{xorg_bin} is suid and readable, using the xpra_Xdummy wrapper")
@@ -266,7 +267,7 @@ def wrap_cmd_str(cmd) -> str:
         while cmd:
             item = cmd[0]
             l = len(item)
-            if (item.startswith("-") or item.startswith("+")) and len(cmd)>1:
+            if (item.startswith("-") or item.startswith("+")) and len(cmd) > 1:
                 l += len(cmd[1])
             if s and len(s)+l > 55:
                 break
@@ -297,7 +298,8 @@ def get_build_info() -> list[str]:
         warn(f"Error: could not find the source information: {e}")
     try:
         from xpra.build_info import BUILD_TYPE
-        info.append(f"{BUILD_TYPE} build")
+        if BUILD_TYPE:
+            info.append(f"{BUILD_TYPE} build")
     except ImportError:
         pass
     try:
@@ -334,7 +336,7 @@ def name_to_field(name: str) -> str:
     return name.replace("-", "_")
 
 
-def save_config(conf_file: str, config, keys, extras_types=None):
+def save_config(conf_file: str, config, keys, extras_types=None) -> None:
     with open(conf_file, "w", encoding="utf8") as f:
         option_types = OPTION_TYPES.copy()
         if extras_types:
@@ -355,7 +357,7 @@ def read_config(conf_file: str) -> dict[str, Any]:
         If the same key is specified more than once,
         the value for this key will be an array of strings.
     """
-    d : dict[str, str | list[str]] = {}
+    d: dict[str, str | list[str]] = {}
     if not os.path.exists(conf_file) or not os.path.isfile(conf_file):
         debug("read_config(%s) is not a file or does not exist", conf_file)
         return d
@@ -460,7 +462,7 @@ def read_xpra_conf(conf_dir: str) -> dict[str, Any]:
     return d
 
 
-def read_xpra_defaults(username: str | None = None, uid=None, gid=None):
+def read_xpra_defaults(username: str | None = None, uid=None, gid=None) -> dict[str, Any]:
     """
         Reads the global <xpra_conf_filename> from the <conf_dir>
         and then the user-specific one.
@@ -469,7 +471,7 @@ def read_xpra_defaults(username: str | None = None, uid=None, gid=None):
         If the <conf_dir> is not specified, we figure out its location.
     """
     dirs = get_xpra_defaults_dirs(username, uid, gid)
-    defaults = {}
+    defaults: dict[str, Any] = {}
     for d in dirs:
         conf_data = read_xpra_conf(d)
         defaults.update(conf_data)
@@ -478,7 +480,7 @@ def read_xpra_defaults(username: str | None = None, uid=None, gid=None):
     return defaults
 
 
-def get_xpra_defaults_dirs(username: str | None = None, uid=None, gid=None):
+def get_xpra_defaults_dirs(username: str | None = None, uid=None, gid=None) -> list[str]:
     from xpra.platform.paths import get_default_conf_dirs, get_system_conf_dirs, get_user_conf_dirs
     # load config files in this order (the later ones override earlier ones):
     # * application defaults, ie:
@@ -505,26 +507,21 @@ def get_xpra_defaults_dirs(username: str | None = None, uid=None, gid=None):
     return defaults_dirs
 
 
-def may_create_user_config(xpra_conf_filename:str = DEFAULT_XPRA_CONF_FILENAME):
+def may_create_user_config(xpra_conf_filename: str = DEFAULT_XPRA_CONF_FILENAME) -> None:
     from xpra.platform.paths import get_user_conf_dirs
     # save a user config template:
     udirs = get_user_conf_dirs()
-    if udirs:
-        has_user_conf = None
-        for d in udirs:
-            if os.path.exists(d):
-                has_user_conf = d
-                break
-        if not has_user_conf:
-            debug("no user configuration file found, trying to create one")
-            for d in udirs:
-                ad = os.path.expanduser(d)
-                conf_file = os.path.join(ad, xpra_conf_filename)
-                try:
-                    if not os.path.exists(ad):
-                        os.makedirs(ad, int('700', 8))
-                    with open(conf_file, 'w', encoding="utf8") as f:
-                        f.write("""# xpra user configuration file
+    if any(os.path.exists(osexpand(d)) for d in udirs):
+        return
+    debug("no user configuration file found, trying to create one")
+    for d in udirs:
+        ad = os.path.expanduser(d)
+        conf_file = os.path.join(ad, xpra_conf_filename)
+        try:
+            if not os.path.exists(ad):
+                os.makedirs(ad, int('700', 8))
+            with open(conf_file, 'w', encoding="utf8") as f:
+                f.write("""# xpra user configuration file
 # place your custom settings in this file
 # they will take precedence over the system default ones.
 
@@ -536,16 +533,16 @@ def may_create_user_config(xpra_conf_filename:str = DEFAULT_XPRA_CONF_FILENAME):
 # see the xpra manual at:
 # https://github.com/Xpra-org/xpra/tree/master/docs/
 """)
-                        f.flush()
-                    debug(f"created default config in {d!r}")
-                    break
-                except Exception as e:
-                    debug(f"failed to create default config in {conf_file!r}: {e}")
+                f.flush()
+            debug(f"created default config in {d!r}")
+            return
+        except Exception as e:
+            debug(f"failed to create default config in {conf_file!r}: {e}")
 
 
-OPTIONS_VALIDATION : dict[str, Callable] = {}
+OPTIONS_VALIDATION: dict[str, Callable] = {}
 
-OPTION_TYPES = {
+OPTION_TYPES: dict[str, Any] = {
     # string options:
     "encoding"          : str,
     "opengl"            : str,
@@ -781,19 +778,19 @@ BIND_OPTIONS : list[str] = [
 
 # keep track of the options added since v5,
 # so we can generate command lines that work with older supported versions:
-OPTIONS_ADDED_SINCE_V5 : list[str] = [
+OPTIONS_ADDED_SINCE_V5: list[str] = [
     "minimal",
 ]
-OPTIONS_COMPAT_NAMES : dict[str, str] = {
-    "--compression_level=" : "-z"
+OPTIONS_COMPAT_NAMES: dict[str, str] = {
+    "--compression_level=": "-z"
 }
 
-CLIENT_OPTIONS : list[str] = [
+CLIENT_OPTIONS: list[str] = [
     "title", "username", "password", "session-name",
     "dock-icon", "tray-icon", "window-icon",
     "clipboard", "clipboard-direction", "clipboard-filter-file",
     "remote-clipboard", "local-clipboard",
-    "tcp-encryption", "tcp-encryption-keyfile", "encryption",  "encryption-keyfile",
+    "tcp-encryption", "tcp-encryption-keyfile", "encryption", "encryption-keyfile",
     "systemd-run", "systemd-run-args",
     "socket-dir", "socket-dirs", "client-socket-dirs",
     "border", "window-close", "min-size", "max-size", "desktop-scaling",
@@ -822,9 +819,10 @@ CLIENT_OPTIONS : list[str] = [
     "compressors", "packet-encoders",
     "key-shortcut",
     "env",
+    "ssh",
 ]
 
-CLIENT_ONLY_OPTIONS : list[str] = [
+CLIENT_ONLY_OPTIONS: list[str] = [
     "username", "swap-keys", "dock-icon",
     "tray", "delay-tray", "tray-icon",
     "attach",
@@ -833,7 +831,7 @@ CLIENT_ONLY_OPTIONS : list[str] = [
 
 # options that clients can pass to the proxy
 # and which will be forwarded to the new proxy instance process:
-PROXY_START_OVERRIDABLE_OPTIONS : list[str] = [
+PROXY_START_OVERRIDABLE_OPTIONS: list[str] = [
     "env", "start-env", "chdir",
     "dpi",
     "encoding", "encodings",
@@ -878,35 +876,35 @@ del tmp
 
 def get_default_key_shortcuts() -> list[str]:
     return [shortcut for e, shortcut in (
-        (True,   "Control+Menu:toggle_keyboard_grab"),
-        (True,   "Shift+Menu:toggle_pointer_grab"),
+        (True, "Control+Menu:toggle_keyboard_grab"),
+        (True, "Shift+Menu:toggle_pointer_grab"),
         (not OSX, "Shift+F11:toggle_fullscreen"),
-        (OSX,    "Control+F11:toggle_fullscreen"),
-        (True,   "#+F1:show_menu"),
-        (True,   "#+F2:show_start_new_command"),
-        (True,   "#+F3:show_bug_report"),
-        (True,   "#+F4:quit"),
-        (True,   "#+F5:show_window_info"),
-        (True,   "#+F6:show_shortcuts"),
-        (True,   "#+F7:show_docs"),
-        (True,   "#+F8:toggle_keyboard_grab"),
-        (True,   "#+F9:toggle_pointer_grab"),
-        (True,   "#+F10:magic_key"),
-        (True,   "#+F11:show_session_info"),
-        (True,   "#+F12:toggle_debug"),
-        (True,   "#+plus:scaleup"),
-        (OSX,    "#+plusminus:scaleup"),
-        (True,   "#+minus:scaledown"),
-        (True,   "#+underscore:scaledown"),
-        (OSX,    "#+emdash:scaledown"),
-        (True,   "#+KP_Add:scaleup"),
-        (True,   "#+KP_Subtract:scaledown"),
-        (True,   "#+KP_Multiply:scalereset"),
-        (True,   "#+bar:scalereset"),
-        (True,   "#+question:scalingoff"),
-        (OSX,    "#+degree:scalereset"),
-        (OSX,    "meta+grave:void"),
-        (OSX,    "meta+shift+asciitilde:void"),
+        (OSX, "Control+F11:toggle_fullscreen"),
+        (True, "#+F1:show_menu"),
+        (True, "#+F2:show_start_new_command"),
+        (True, "#+F3:show_bug_report"),
+        (True, "#+F4:quit"),
+        (True, "#+F5:show_window_info"),
+        (True, "#+F6:show_shortcuts"),
+        (True, "#+F7:show_docs"),
+        (True, "#+F8:toggle_keyboard_grab"),
+        (True, "#+F9:toggle_pointer_grab"),
+        (True, "#+F10:magic_key"),
+        (True, "#+F11:show_session_info"),
+        (True, "#+F12:toggle_debug"),
+        (True, "#+plus:scaleup"),
+        (OSX, "#+plusminus:scaleup"),
+        (True, "#+minus:scaledown"),
+        (True, "#+underscore:scaledown"),
+        (OSX, "#+emdash:scaledown"),
+        (True, "#+KP_Add:scaleup"),
+        (True, "#+KP_Subtract:scaledown"),
+        (True, "#+KP_Multiply:scalereset"),
+        (True, "#+bar:scalereset"),
+        (True, "#+question:scalingoff"),
+        (OSX, "#+degree:scalereset"),
+        (OSX, "meta+grave:void"),
+        (OSX, "meta+shift+asciitilde:void"),
     )
         if e]
 
@@ -976,7 +974,7 @@ def unexpand_all(paths: list[str]) -> list[str]:
     return [unexpand(x) for x in paths]
 
 
-GLOBAL_DEFAULTS = None
+GLOBAL_DEFAULTS: dict[str, Any] | None = None
 # lowest common denominator here
 # (the xpra.conf file shipped is generally better tuned than this - especially for 'xvfb')
 
@@ -1234,10 +1232,10 @@ CLONES: dict[str, str] = {}
 NO_FILE_OPTIONS = ("daemon", )
 
 
-TRUE_OPTIONS: tuple[Any, ...] = ("yes", "true", "1", "on", True)
-FALSE_OPTIONS: tuple[Any, ...] = ("no", "false", "0", "off", False)
-ALL_BOOLEAN_OPTIONS: tuple[Any, ...] = tuple(list(TRUE_OPTIONS)+list(FALSE_OPTIONS))
-OFF_OPTIONS: tuple[str, ...] = ("off", )
+TRUE_OPTIONS: Sequence[str, bool] = ("yes", "true", "1", "on", True)
+FALSE_OPTIONS: Sequence[str, bool] = ("no", "false", "0", "off", False)
+ALL_BOOLEAN_OPTIONS: Sequence[str, bool] = tuple(list(TRUE_OPTIONS)+list(FALSE_OPTIONS))
+OFF_OPTIONS: Sequence[str] = ("off", )
 
 
 def str_to_bool(v: Any, default: bool = True) -> bool:
@@ -1358,7 +1356,7 @@ def do_validate_config(d:dict, discard, extras_types:dict, extras_validation:dic
     validations.update(extras_validation)
     option_types = OPTION_TYPES.copy()
     option_types.update(extras_types)
-    nd : dict[str, Any] = {}
+    nd: dict[str, Any] = {}
     for k, v in d.items():
         if k in discard:
             warn(f"Warning: option {k!r} is not allowed in configuration files")
@@ -1369,23 +1367,23 @@ def do_validate_config(d:dict, discard, extras_types:dict, extras_validation:dic
                 continue
             warn(f"Warning: invalid option: {k!r}")
             continue
-        if vt == str:
+        if vt is str:
             if not isinstance(v, str):
                 warn(f"invalid value for {k!r}: {type(v)} (string required)")
                 continue
-        elif vt == int:
+        elif vt is int:
             v = parse_bool_or_number(int, k, v)
             if v is None:
                 continue
-        elif vt == float:
+        elif vt is float:
             v = parse_number(float, k, v)
             if v is None:
                 continue
-        elif vt == bool:
+        elif vt is bool:
             v = parse_bool_or(k, v)
             if v is None:
                 continue
-        elif vt == list:
+        elif vt is list:
             if isinstance(v, str):
                 # could just be that we specified it only once..
                 v = [v]
@@ -1423,16 +1421,16 @@ def make_defaults_struct(extras_defaults=None, extras_types=None, extras_validat
                                    username, uid, gid)
 
 
-def do_make_defaults_struct(extras_defaults:dict, extras_types:dict, extras_validation:dict,
-                            username:str, uid:int, gid:int) -> XpraConfig:
-    #populate config with default values:
+def do_make_defaults_struct(extras_defaults: dict, extras_types: dict, extras_validation: dict,
+                            username:str, uid: int, gid: int) -> XpraConfig:
+    # populate config with default values:
     if not username and uid:
         username = get_username_for_uid(uid)
     defaults = read_xpra_defaults(username, uid, gid)
     return dict_to_validated_config(defaults, extras_defaults, extras_types, extras_validation)
 
 
-def dict_to_validated_config(d:dict, extras_defaults=None, extras_types=None, extras_validation=None) -> XpraConfig:
+def dict_to_validated_config(d: dict, extras_defaults=None, extras_types=None, extras_validation=None) -> XpraConfig:
     options = get_defaults().copy()
     if extras_defaults:
         options.update(extras_defaults)
@@ -1452,7 +1450,7 @@ def dict_to_config(options) -> XpraConfig:
     return config
 
 
-def fixup_debug_option(value:str) -> str:
+def fixup_debug_option(value: str) -> str:
     """ backwards compatible parsing of the debug option, which used to be a boolean """
     if not value:
         return ""
@@ -1477,7 +1475,7 @@ def csvstrl(value) -> str:
     return csvstr(value).lower()
 
 
-def nodupes(s):
+def nodupes(s) -> list[str]:
     return remove_dupes(x.strip().lower() for x in s.split(","))
 
 
@@ -1652,8 +1650,8 @@ def abs_paths(options) -> None:
               "ssl-key", "ssl-cert", "ssl-ca-certs"):
         f = k.replace("-", "_")
         v = getattr(options, f)
-        if v and (k!="ssl-ca-certs" or v!="default"):
-            if os.path.isabs(v) or v=="auto":
+        if v and (k != "ssl-ca-certs" or v != "default"):
+            if os.path.isabs(v) or v == "auto":
                 continue
             if v.startswith("~") or v.startswith("$"):
                 continue
@@ -1675,32 +1673,30 @@ def fixup_options(options) -> None:
         options.remote_xpra = [options.remote_xpra]
 
 
-def main():
+def main(argv):
     from xpra.util.str_fn import nonl
 
     def print_options(o):
-        for k,ot in sorted(OPTION_TYPES.items()):
+        for k, ot in sorted(OPTION_TYPES.items()):
             v = getattr(o, name_to_field(k), "")
-            if ot==bool and v is None:
+            if ot is bool and v is None:
                 v = "Auto"
             if isinstance(v, list):
                 v = csv(str(x) for x in v)
             print("* %-32s : %s" % (k, nonl(v)))
     from xpra.platform import program_context
-    from xpra.log import enable_color
+    from xpra.log import enable_color, consume_verbose_argv
     with program_context("Config-Info", "Config Info"):
         enable_color()
-        args = list(sys.argv[1:])
-        if "-v" in args or "--verbose" in sys.argv:
+        if consume_verbose_argv(argv, "all"):
             global debug
 
             def print_debug(*args):
                 print(args[0] % args[1:])
             debug = print_debug
-            args.remove("-v")
 
-        if args:
-            for filename in args:
+        if len(argv) > 1:
+            for filename in argv[1:]:
                 print("")
                 print(f"Configuration file {filename!r}")
                 if not os.path.exists(filename):
@@ -1715,4 +1711,4 @@ def main():
 
 
 if __name__ == "__main__":
-    main()
+    main(sys.argv)
