@@ -21,15 +21,17 @@ from xpra.codecs.image import ImageWrapper
 from xpra.codecs.video import getVideoHelper, PREFERRED_ENCODER_ORDER
 from xpra.scripts.config import parse_number, str_to_bool
 from xpra.common import FULL_INFO, ConnectionMessage
-from xpra.os_util import get_hex_uuid
+from xpra.os_util import get_hex_uuid, gi_import
 from xpra.util.objects import typedict
-from xpra.util.str_fn import csv, ellipsizer, strtobytes, nicestr
+from xpra.util.str_fn import csv, Ellipsizer, strtobytes, nicestr
 from xpra.util.env import envint, envbool, first_time
 from xpra.util.version import XPRA_VERSION, vparts
 from xpra.util.thread import start_thread
 from xpra.server.core import get_server_info, get_thread_info, proto_crypto_caps
 from xpra.log import Logger
 
+GLib = gi_import("GLib")
+
 log = Logger("proxy")
 enclog = Logger("encoding")
 
@@ -51,6 +53,13 @@ def number(k, v):
     return parse_number(int, k, v)
 
 
+SESSION_OPTION_WHITELIST: dict[str, Callable] = {
+    "compression_level": number,
+    "lz4": str_to_bool,
+    "rencodeplus": str_to_bool,
+}
+
+
 # noinspection PyMethodMayBeStatic
 class ProxyInstance:
 
@@ -69,7 +78,7 @@ class ProxyInstance:
             session_options,
             video_encoder_modules,
             disp_desc, cipher, cipher_mode, encryption_key,
-            "%s: %s.." % (type(caps), ellipsizer(caps))))
+            "%s: %s.." % (type(caps), Ellipsizer(caps))))
         self.uuid = get_hex_uuid()
         self.client_protocol = None
         self.client_has_more = False
@@ -87,11 +96,11 @@ class ProxyInstance:
         self.client_challenge_packet: tuple = ()
         self.exit = False
         self.lost_windows = set()
-        self.encode_queue = SimpleQueue()  # holds draw packets to encode
+        self.encode_queue: SimpleQueue[PacketType] = SimpleQueue()  # holds draw packets to encode
         self.encode_thread: Thread | None = None
         # setup protocol wrappers:
-        self.server_packets: Queue[tuple] = Queue(PROXY_QUEUE_SIZE)
-        self.client_packets: Queue[tuple] = Queue(PROXY_QUEUE_SIZE)
+        self.server_packets: Queue[PacketType] = Queue(PROXY_QUEUE_SIZE)
+        self.client_packets: Queue[PacketType] = Queue(PROXY_QUEUE_SIZE)
         self.video_encoding_defs: dict[str, dict[str, list]] = {}
         self.video_encoders: dict[int, Any] = {}
         self.video_encoders_dst_formats = ()
@@ -230,7 +239,7 @@ class ProxyInstance:
 
     ################################################################################
 
-    def send_hello(self, challenge_response=None, client_salt=None) -> None:
+    def send_hello(self, challenge_response=b"", client_salt=b"") -> None:
         hello = self.filter_client_caps(self.caps)
         if challenge_response:
             hello.update({
@@ -242,14 +251,8 @@ class ProxyInstance:
 
     def sanitize_session_options(self, options) -> dict[str, Any]:
         d = {}
-
-        OPTION_WHITELIST: dict[str, Callable] = {
-            "compression_level": number,
-            "lz4": str_to_bool,
-            "rencodeplus": str_to_bool,
-        }
         for k, v in options.items():
-            parser = OPTION_WHITELIST.get(k)
+            parser = SESSION_OPTION_WHITELIST.get(k)
             if parser:
                 log("trying to add %s=%s using %s", k, v, parser)
                 try:
@@ -302,12 +305,12 @@ class ProxyInstance:
         self.client_packets.put(packet)
         self.client_protocol.source_has_more()
 
-    def get_client_packet(self):
+    def get_client_packet(self) -> tuple[PacketType, bool, bool]:
         # server wants a packet
         p = self.client_packets.get()
         s = self.client_packets.qsize()
         log("sending to client: %s (queue size=%i)", p[0], s)
-        return p, None, None, None, True, s > 0 or self.server_has_more
+        return p, True, s > 0 or self.server_has_more
 
     def process_client_packet(self, proto, packet: PacketType) -> None:
         packet_type = str(packet[0])
@@ -354,7 +357,7 @@ class ProxyInstance:
 
     def replace_packet_item(self, packet: PacketType,
                             index: int,
-                            new_value: Compressed | str | dict) -> PacketType:
+                            new_value: Compressed | str | dict | int) -> PacketType:
         # make the packet data mutable and replace the contents at `index`:
         assert index > 0
         lpacket = list(packet)
@@ -372,7 +375,7 @@ class ProxyInstance:
         p = self.server_packets.get()
         s = self.server_packets.qsize()
         log("sending to server: %s (queue size=%i)", p[0], s)
-        return p, None, None, None, True, s > 0 or self.client_has_more
+        return p, True, s > 0 or self.client_has_more
 
     def _packet_recompress(self, packet: PacketType, index: int, name: str) -> PacketType:
         if len(packet) <= index:
@@ -534,7 +537,7 @@ class ProxyInstance:
                 # client may have already responded to the challenge,
                 # so we have to handle authentication from this end
                 server_salt = strtobytes(packet[1])
-                l = len(server_salt)
+                length = len(server_salt)
                 digest = str(packet[3])
                 salt_digest = "xor"
                 if len(packet) >= 5:
@@ -544,14 +547,14 @@ class ProxyInstance:
                     return
                 if salt_digest == "xor":
                     # with xor, we have to match the size
-                    if l < 16:
+                    if length < 16:
                         raise ValueError("server salt is too short: only {l} bytes, minimum is 16")
-                    if l > 256:
+                    if length > 256:
                         raise ValueError("server salt is too long: {l} bytes, maximum is 256")
                 else:
                     # other digest, 32 random bytes is enough:
-                    l = 32
-                client_salt = get_salt(l)
+                    length = 32
+                client_salt = get_salt(length)
                 salt = gendigest(salt_digest, client_salt, server_salt)
                 challenge_response = gendigest(digest, password, salt)
                 if not challenge_response:
@@ -567,9 +570,9 @@ class ProxyInstance:
         # empty the encode queue:
         q = self.encode_queue
         if q:
-            q.put_nowait(None)
+            q.put_nowait(())
             q = SimpleQueue()
-            q.put(None)
+            q.put(())
             self.encode_queue = q
 
     def delvideo(self, wid: int):
@@ -583,10 +586,10 @@ class ProxyInstance:
 
         while not self.exit:
             packet = self.encode_queue.get()
-            if packet is None:
+            if not packet:
                 return
+            packet_type = str(packet[0])
             try:
-                packet_type = str(packet[0])
                 if packet_type == "lost-window":
                     wid = packet[1]
                     self.lost_windows.remove(wid)
@@ -606,8 +609,8 @@ class ProxyInstance:
                         self.delvideo(wid)
                 else:
                     enclog.warn("unexpected encode packet: %s", packet_type)
-            except Exception:
-                enclog.warn("error encoding packet", exc_info=True)
+            except (ValueError, RuntimeError):
+                enclog.warn(f"Warning: failed encoding packet {packet_type!r}", exc_info=True)
 
     def process_draw(self, packet: PacketType) -> None:
         wid, x, y, width, height, encoding, pixels, _, rowstride, client_options = packet[1:11]
@@ -646,6 +649,7 @@ class ProxyInstance:
             updated = packet
             if strip_alpha:
                 # passthrough as plain RGB:
+                # noinspection PyPep8Naming
                 Xindex = rgb_format.upper().find("X")
                 if Xindex >= 0 and len(rgb_format) == 4:
                     # force clear alpha (which may be garbage):
@@ -745,7 +749,11 @@ class ProxyInstance:
             self.video_encoders_last_used_time[wid] = monotonic()  # just to make sure this is always set
         # actual video compression:
         enclog("proxy compression using %s with quality=%s, speed=%s", ve, quality, speed)
-        data, out_options = ve.compress_image(image, quality, speed, encoder_options)
+        if quality > 0:
+            encoder_options["quality"] = quality
+        if speed > 0:
+            encoder_options["speed"] = speed
+        data, out_options = ve.compress_image(image, encoder_options)
         # pass through some options if we don't have them from the encoder
         # (maybe we should also use the "pts" from the real server?)
         for k in ("timestamp", "rgb_format", "depth", "csc"):
@@ -766,7 +774,7 @@ class ProxyInstance:
             idle_time = int(now - vetime)
             enclog("timeout_video_encoders() wid=%s, idle_time=%s", wid, idle_time)
             if idle_time and idle_time > VIDEO_TIMEOUT:
-                self.encode_queue.put(["check-video-timeout", wid])
+                self.encode_queue.put(("check-video-timeout", wid))
         return True  # run again
 
     def _find_video_encoder(self, video_encoding, rgb_format):
