@@ -7,6 +7,7 @@ import os
 from subprocess import Popen
 from shutil import which
 from typing import Any
+from collections.abc import Sequence, Iterable
 
 from xpra.net.compression import Compressed
 from xpra.server.source.stub_source_mixin import StubSourceMixin
@@ -17,6 +18,8 @@ from xpra.util.str_fn import csv, bytestostr
 from xpra.util.env import envint, envbool, first_time
 from xpra.log import Logger
 
+GLib = gi_import("GLib")
+
 log = Logger("audio")
 
 NEW_STREAM_SOUND = envbool("XPRA_NEW_STREAM_SOUND", True)
@@ -27,10 +30,12 @@ class FakeSink:
     def __init__(self, codec):
         self.codec = codec
 
-    def add_data(self, *args):
+    @staticmethod
+    def add_data(*args):
         log("FakeSink.add_data%s ignored", args)
 
-    def cleanup(self, *args):
+    @staticmethod
+    def cleanup(*args):
         log("FakeSink.cleanup%s ignored", args)
 
 
@@ -42,7 +47,7 @@ def stop_proc(proc) -> None:
         return
     try:
         proc.terminate()
-    except Exception:
+    except OSError:
         log("failed to stop subprocess %s", proc)
 
 
@@ -80,8 +85,8 @@ class AudioMixin(StubSourceMixin):
         self.pulseaudio_id = ""
         self.pulseaudio_cookie_hash = ""
         self.pulseaudio_server = ""
-        self.audio_decoders: tuple[str, ...] = ()
-        self.audio_encoders: tuple[str, ...] = ()
+        self.audio_decoders: Sequence[str] = ()
+        self.audio_encoders: Sequence[str] = ()
         self.audio_receive = False
         self.audio_send = False
         self.audio_fade_timer = 0
@@ -101,12 +106,12 @@ class AudioMixin(StubSourceMixin):
         for proc, timer in timers.items():
             timer = self.new_stream_timers.pop(proc, 0)
             if timer:
-                self.source_remove(timer)
+                GLib.source_remove(timer)
             stop_proc(proc)
 
     def parse_client_caps(self, c: typedict) -> None:
-        self.wants_audio = "audio" in c.strtupleget("wants")
         audio = typedict(c.dictget("audio") or {})
+        self.wants_audio = "audio" in c.strtupleget("wants") or audio.boolget("send") or audio.boolget("receive")
         if audio:
             self.pulseaudio_id = audio.strget("pulseaudio.id")
             self.pulseaudio_cookie_hash = audio.strget("pulseaudio.cookie-hash")
@@ -185,8 +190,7 @@ class AudioMixin(StubSourceMixin):
             log.warn(" %s", x)
         return False
 
-    def start_sending_audio(self, codec: str = "", volume: float = 1.0,
-                            new_stream=None, new_buffer=None, skip_client_codec_check=False):
+    def start_sending_audio(self, codec: str = "", volume: float = 1.0):
         log("start_sending_audio(%s)", codec)
         ss = None
         if getattr(self, "suspended", False):
@@ -210,7 +214,7 @@ class AudioMixin(StubSourceMixin):
         elif codec not in self.speaker_codecs:
             log.warn("Warning: invalid codec specified: %s", codec)
             return None
-        elif (codec not in self.audio_decoders) and not skip_client_codec_check:
+        elif codec not in self.audio_decoders:
             log.warn("Error sending audio: invalid codec '%s'", codec)
             log.warn(" is not in the list of decoders supported by the client: %s", csv(self.audio_decoders))
             return None
@@ -218,17 +222,17 @@ class AudioMixin(StubSourceMixin):
             return None
         try:
             from xpra.audio.wrapper import start_sending_audio  # pylint: disable=import-outside-toplevel
-            plugins = self.audio_properties.strtupleget("plugins")
+            plugins = self.audio_properties.strtupleget("sources")
             ss = start_sending_audio(plugins, self.audio_source_plugin,
-                                     None, codec, volume, True, [codec],
+                                     "", codec, volume, True, [codec],
                                      self.pulseaudio_server, self.pulseaudio_id)
             self.audio_source = ss
             log("start_sending_audio() audio source=%s", ss)
             if not ss:
                 return None
             ss.sequence = self.audio_source_sequence
-            ss.connect("new-buffer", new_buffer or self.new_audio_buffer)
-            ss.connect("new-stream", new_stream or self.new_stream)
+            ss.connect("new-buffer", self.new_audio_buffer)
+            ss.connect("new-stream", self.new_stream)
             ss.connect("info", self.audio_source_info)
             ss.connect("exit", self.audio_source_exit)
             ss.connect("error", self.audio_source_error)
@@ -244,7 +248,7 @@ class AudioMixin(StubSourceMixin):
                 # tell the client we're not sending anything:
                 self.send_eos(codec)
 
-    def audio_source_error(self, source, message) -> None:
+    def audio_source_error(self, source, message: str) -> None:
         # this should be printed to stderr by the audio process already
         if source == self.audio_source:
             log("audio capture error: %s", message)
@@ -254,7 +258,8 @@ class AudioMixin(StubSourceMixin):
         if source == self.audio_source:
             self.stop_sending_audio()
 
-    def audio_source_info(self, source, info) -> None:
+    @staticmethod
+    def audio_source_info(source, info: dict) -> None:
         log("audio_source_info(%s, %s)", source, info)
 
     def stop_sending_audio(self) -> None:
@@ -270,7 +275,7 @@ class AudioMixin(StubSourceMixin):
     def send_eos(self, codec: str, sequence: int = 0) -> None:
         log("send_eos(%s, %s)", codec, sequence)
         # tell the client this is the end:
-        self.send_more("sound-data", codec, "",
+        self.send_more("sound-data", codec, b"",
                        {
                            "end-of-stream": True,
                            "sequence": sequence,
@@ -299,11 +304,11 @@ class AudioMixin(StubSourceMixin):
             from xpra.util.child_reaper import getChildReaper  # pylint: disable=import-outside-toplevel
             getChildReaper().add_process(proc, "new-stream-sound", cmd, ignore=True, forget=True)
 
-            def stop_new_stream_notification():
+            def stop_new_stream_notification() -> None:
                 if self.new_stream_timers.pop(proc, None):
                     stop_proc(proc)
 
-            timer = self.timeout_add(NEW_STREAM_SOUND_STOP * 1000, stop_new_stream_notification)
+            timer = GLib.timeout_add(NEW_STREAM_SOUND_STOP * 1000, stop_new_stream_notification)
             self.new_stream_timers[proc] = timer
         except Exception as e:
             log("new_stream_sound() error playing new stream sound", exc_info=True)
@@ -311,16 +316,16 @@ class AudioMixin(StubSourceMixin):
             log.error(f" using: {cmd_str}:")
             log.estr(e)
 
-    def new_stream(self, audio_source, codec: str) -> None:
+    def new_stream(self, audio_source, codec: str) -> bool:
         log("new_stream(%s, %s)", audio_source, codec)
         self.new_stream_sound()
         if self.audio_source != audio_source:
             log("dropping new-stream signal (current source=%s, signal source=%s)", self.audio_source, audio_source)
-            return
+            return False
         codec = codec or audio_source.codec
         audio_source.codec = codec
         # tell the client this is the start:
-        self.send("sound-data", codec, "", {
+        self.send("sound-data", codec, b"", {
             "start-of-stream": True,
             "codec": codec,
             "sequence": audio_source.sequence,
@@ -328,8 +333,8 @@ class AudioMixin(StubSourceMixin):
         self.call_update_av_sync_delay()
         # run it again after 10 seconds,
         # by that point the source info will actually be populated:
-        GLib = gi_import("GLib")
         GLib.timeout_add(10 * 1000, self.call_update_av_sync_delay)
+        return False
 
     def call_update_av_sync_delay(self) -> None:
         # loose coupling with avsync mixin:
@@ -338,7 +343,7 @@ class AudioMixin(StubSourceMixin):
         if callable(update_av_sync):
             update_av_sync()  # pylint: disable=not-callable
 
-    def new_audio_buffer(self, audio_source, data, metadata, packet_metadata=None) -> None:
+    def new_audio_buffer(self, audio_source, data: bytes, metadata: dict, packet_metadata: Iterable = ()) -> None:
         log("new_audio_buffer(%s, %s, %s, %s) info=%s",
             audio_source, len(data or []), metadata, [len(x) for x in packet_metadata], audio_source.info)
         if self.audio_source != audio_source or self.is_closed():
@@ -351,24 +356,15 @@ class AudioMixin(StubSourceMixin):
         if packet_metadata:
             # the packet metadata is compressed already:
             packet_metadata = Compressed("packet metadata", packet_metadata, can_inline=True)
-        # don't drop the first 10 buffers
-        can_drop_packet = (audio_source.info or {}).get("buffer_count", 0) > 10
-        self.send_audio_data(audio_source, data, metadata, packet_metadata, can_drop_packet)
+        self.send_audio_data(audio_source, data, metadata, packet_metadata)
 
-    def send_audio_data(self, audio_source, data, metadata, packet_metadata=None, can_drop_packet=False) -> None:
-        packet_data = [audio_source.codec, Compressed(audio_source.codec, data), metadata, packet_metadata or ()]
+    def send_audio_data(self, audio_source, data: bytes, metadata: dict,
+                        packet_metadata: Iterable = ()) -> None:
+        packet_data = [audio_source.codec, Compressed(audio_source.codec, data, True), metadata, packet_metadata or ()]
         sequence = audio_source.sequence
         if sequence >= 0:
             metadata["sequence"] = sequence
-        fail_cb = None
-        if can_drop_packet:
-            def audio_data_fail_cb():
-                # ideally we would tell gstreamer to send an audio "key frame"
-                # or synchronization point to ensure the stream recovers
-                log("a audio data buffer was not received and will not be resent")
-
-            fail_cb = audio_data_fail_cb
-        self.send("sound-data", *packet_data, synchronous=False, fail_cb=fail_cb, will_have_more=True)
+        self.send("sound-data", *packet_data, synchronous=False, will_have_more=True)
 
     def stop_receiving_audio(self) -> None:
         ss = self.audio_sink
@@ -379,8 +375,7 @@ class AudioMixin(StubSourceMixin):
 
     ##########################################################################
     # audio control commands:
-    def audio_control(self, action, *args):
-        action = bytestostr(action)
+    def audio_control(self, action: str, *args):
         fn = "audio_control_" + action.replace("-", "_")
         method = getattr(self, fn, None)
         log(f"audio_control({action}, {args}) {self}.{fn}={method}")
@@ -415,7 +410,7 @@ class AudioMixin(StubSourceMixin):
         step = 1.0 / (delay / 100.0)
         log("audio_control fadein delay=%s, step=%1.f", delay, step)
 
-        def fadein():
+        def fadein() -> bool:
             ss = self.audio_source
             if not ss:
                 return False
@@ -427,7 +422,7 @@ class AudioMixin(StubSourceMixin):
             return volume < 1.0
 
         self.cancel_audio_fade_timer()
-        self.audio_fade_timer = self.timeout_add(100, fadein)
+        self.audio_fade_timer = GLib.timeout_add(100, fadein)
         return "fadein started"
 
     def audio_control_start(self, codec: str = "") -> str:
@@ -452,7 +447,7 @@ class AudioMixin(StubSourceMixin):
         step = 1.0 / (delay / 100.0)
         log("audio_control fadeout delay=%s, step=%1.f", delay, step)
 
-        def fadeout():
+        def fadeout() -> bool:
             ss = self.audio_source
             if not ss:
                 return False
@@ -465,7 +460,7 @@ class AudioMixin(StubSourceMixin):
             return False
 
         self.cancel_audio_fade_timer()
-        self.audio_fade_timer = self.timeout_add(100, fadeout)
+        self.audio_fade_timer = GLib.timeout_add(100, fadeout)
         return "fadeout started"
 
     def audio_control_new_sequence(self, seq_str) -> str:
@@ -476,9 +471,9 @@ class AudioMixin(StubSourceMixin):
         sft = self.audio_fade_timer
         if sft:
             self.audio_fade_timer = 0
-            self.source_remove(sft)
+            GLib.source_remove(sft)
 
-    def audio_data(self, codec, data, metadata, packet_metadata=()) -> None:
+    def audio_data(self, codec: str, data: bytes, metadata: dict, packet_metadata: Iterable = ()) -> None:
         log("audio_data(%s, %s, %s, %s) audio sink=%s",
             codec, len(data or []), metadata, packet_metadata, self.audio_sink)
         if self.is_closed():
@@ -498,7 +493,7 @@ class AudioMixin(StubSourceMixin):
                 self.audio_sink = FakeSink(codec)
                 return
             try:
-                def audio_sink_error(*args):
+                def audio_sink_error(*args) -> None:
                     log("audio_sink_error%s", args)
                     log.warn("Warning: stopping audio input because of an error")
                     self.stop_receiving_audio()
@@ -548,7 +543,7 @@ class AudioMixin(StubSourceMixin):
         return {"audio": self.get_audio_info()}
 
     def get_audio_info(self) -> dict[str, Any]:
-        def audio_info(supported, subprocess_wrapper, codecs):
+        def audio_info(supported, subprocess_wrapper, codecs) -> dict[str, Any]:
             i = {"codecs": codecs}
             if not supported:
                 i["state"] = "disabled"
