@@ -15,6 +15,7 @@ from xpra.log import Logger
 
 RandR = RandRBindings()
 
+GLib = gi_import("GLib")
 GObject = gi_import("GObject")
 
 log = Logger("server")
@@ -41,7 +42,7 @@ class XpraDesktopServer(DesktopServerBase):
         super().server_init()
         from xpra.x11.vfb_util import set_initial_resolution, get_desktop_vfb_resolutions
         screenlog(f"server_init() randr={self.randr}, initial-resolutions={self.initial_resolutions}")
-        if not self.randr or self.initial_resolutions == () or not features.display:
+        if not features.display:
             return
         res = self.initial_resolutions or get_desktop_vfb_resolutions(default_refresh_rate=self.refresh_rate)
         if len(res) > 1:
@@ -85,7 +86,7 @@ class XpraDesktopServer(DesktopServerBase):
         # at the same time as he resizes the window..
         self.resize_value = (w, h)
         if not self.resize_timer:
-            self.resize_timer = self.timeout_add(250, self.do_resize)
+            self.resize_timer = GLib.timeout_add(250, self.do_resize)
 
     def do_resize(self) -> None:
         self.resize_timer = 0
